# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.usergroups_single_get_response_data_relationships_global_roles import (
        UsergroupsSingleGetResponseDataRelationshipsGlobalRoles,
    )
    from ..models.usergroups_single_get_response_data_relationships_project_roles import (
        UsergroupsSingleGetResponseDataRelationshipsProjectRoles,
    )
    from ..models.usergroups_single_get_response_data_relationships_users import (
        UsergroupsSingleGetResponseDataRelationshipsUsers,
    )


T = TypeVar("T", bound="UsergroupsSingleGetResponseDataRelationships")


@_attrs_define
class UsergroupsSingleGetResponseDataRelationships:
    """
    Attributes:
        global_roles (Union[Unset, UsergroupsSingleGetResponseDataRelationshipsGlobalRoles]):
        project_roles (Union[Unset, UsergroupsSingleGetResponseDataRelationshipsProjectRoles]):
        users (Union[Unset, UsergroupsSingleGetResponseDataRelationshipsUsers]):
    """

    global_roles: Union[
        Unset, "UsergroupsSingleGetResponseDataRelationshipsGlobalRoles"
    ] = UNSET
    project_roles: Union[
        Unset, "UsergroupsSingleGetResponseDataRelationshipsProjectRoles"
    ] = UNSET
    users: Union[
        Unset, "UsergroupsSingleGetResponseDataRelationshipsUsers"
    ] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        global_roles: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.global_roles, Unset):
            global_roles = self.global_roles.to_dict()

        project_roles: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.project_roles, Unset):
            project_roles = self.project_roles.to_dict()

        users: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.users, Unset):
            users = self.users.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if global_roles is not UNSET:
            field_dict["globalRoles"] = global_roles
        if project_roles is not UNSET:
            field_dict["projectRoles"] = project_roles
        if users is not UNSET:
            field_dict["users"] = users

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.usergroups_single_get_response_data_relationships_global_roles import (
            UsergroupsSingleGetResponseDataRelationshipsGlobalRoles,
        )
        from ..models.usergroups_single_get_response_data_relationships_project_roles import (
            UsergroupsSingleGetResponseDataRelationshipsProjectRoles,
        )
        from ..models.usergroups_single_get_response_data_relationships_users import (
            UsergroupsSingleGetResponseDataRelationshipsUsers,
        )

        d = src_dict.copy()
        _global_roles = d.pop("globalRoles", UNSET)
        global_roles: Union[
            Unset, UsergroupsSingleGetResponseDataRelationshipsGlobalRoles
        ]
        if isinstance(_global_roles, Unset):
            global_roles = UNSET
        else:
            global_roles = UsergroupsSingleGetResponseDataRelationshipsGlobalRoles.from_dict(
                _global_roles
            )

        _project_roles = d.pop("projectRoles", UNSET)
        project_roles: Union[
            Unset, UsergroupsSingleGetResponseDataRelationshipsProjectRoles
        ]
        if isinstance(_project_roles, Unset):
            project_roles = UNSET
        else:
            project_roles = UsergroupsSingleGetResponseDataRelationshipsProjectRoles.from_dict(
                _project_roles
            )

        _users = d.pop("users", UNSET)
        users: Union[Unset, UsergroupsSingleGetResponseDataRelationshipsUsers]
        if isinstance(_users, Unset):
            users = UNSET
        else:
            users = (
                UsergroupsSingleGetResponseDataRelationshipsUsers.from_dict(
                    _users
                )
            )

        usergroups_single_get_response_data_relationships_obj = cls(
            global_roles=global_roles,
            project_roles=project_roles,
            users=users,
        )

        usergroups_single_get_response_data_relationships_obj.additional_properties = (
            d
        )
        return usergroups_single_get_response_data_relationships_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.workitem_comments_list_get_response_data_item_relationships_author import (
        WorkitemCommentsListGetResponseDataItemRelationshipsAuthor,
    )
    from ..models.workitem_comments_list_get_response_data_item_relationships_child_comments import (
        WorkitemCommentsListGetResponseDataItemRelationshipsChildComments,
    )
    from ..models.workitem_comments_list_get_response_data_item_relationships_parent_comment import (
        WorkitemCommentsListGetResponseDataItemRelationshipsParentComment,
    )
    from ..models.workitem_comments_list_get_response_data_item_relationships_project import (
        WorkitemCommentsListGetResponseDataItemRelationshipsProject,
    )


T = TypeVar("T", bound="WorkitemCommentsListGetResponseDataItemRelationships")


@_attrs_define
class WorkitemCommentsListGetResponseDataItemRelationships:
    """
    Attributes:
        author (Union[Unset, WorkitemCommentsListGetResponseDataItemRelationshipsAuthor]):
        child_comments (Union[Unset, WorkitemCommentsListGetResponseDataItemRelationshipsChildComments]):
        parent_comment (Union[Unset, WorkitemCommentsListGetResponseDataItemRelationshipsParentComment]):
        project (Union[Unset, WorkitemCommentsListGetResponseDataItemRelationshipsProject]):
    """

    author: Union[
        Unset, "WorkitemCommentsListGetResponseDataItemRelationshipsAuthor"
    ] = UNSET
    child_comments: Union[
        Unset,
        "WorkitemCommentsListGetResponseDataItemRelationshipsChildComments",
    ] = UNSET
    parent_comment: Union[
        Unset,
        "WorkitemCommentsListGetResponseDataItemRelationshipsParentComment",
    ] = UNSET
    project: Union[
        Unset, "WorkitemCommentsListGetResponseDataItemRelationshipsProject"
    ] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        author: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.author, Unset):
            author = self.author.to_dict()

        child_comments: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.child_comments, Unset):
            child_comments = self.child_comments.to_dict()

        parent_comment: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.parent_comment, Unset):
            parent_comment = self.parent_comment.to_dict()

        project: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.project, Unset):
            project = self.project.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if author is not UNSET:
            field_dict["author"] = author
        if child_comments is not UNSET:
            field_dict["childComments"] = child_comments
        if parent_comment is not UNSET:
            field_dict["parentComment"] = parent_comment
        if project is not UNSET:
            field_dict["project"] = project

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.workitem_comments_list_get_response_data_item_relationships_author import (
            WorkitemCommentsListGetResponseDataItemRelationshipsAuthor,
        )
        from ..models.workitem_comments_list_get_response_data_item_relationships_child_comments import (
            WorkitemCommentsListGetResponseDataItemRelationshipsChildComments,
        )
        from ..models.workitem_comments_list_get_response_data_item_relationships_parent_comment import (
            WorkitemCommentsListGetResponseDataItemRelationshipsParentComment,
        )
        from ..models.workitem_comments_list_get_response_data_item_relationships_project import (
            WorkitemCommentsListGetResponseDataItemRelationshipsProject,
        )

        d = src_dict.copy()
        _author = d.pop("author", UNSET)
        author: Union[
            Unset, WorkitemCommentsListGetResponseDataItemRelationshipsAuthor
        ]
        if isinstance(_author, Unset):
            author = UNSET
        else:
            author = WorkitemCommentsListGetResponseDataItemRelationshipsAuthor.from_dict(
                _author
            )

        _child_comments = d.pop("childComments", UNSET)
        child_comments: Union[
            Unset,
            WorkitemCommentsListGetResponseDataItemRelationshipsChildComments,
        ]
        if isinstance(_child_comments, Unset):
            child_comments = UNSET
        else:
            child_comments = WorkitemCommentsListGetResponseDataItemRelationshipsChildComments.from_dict(
                _child_comments
            )

        _parent_comment = d.pop("parentComment", UNSET)
        parent_comment: Union[
            Unset,
            WorkitemCommentsListGetResponseDataItemRelationshipsParentComment,
        ]
        if isinstance(_parent_comment, Unset):
            parent_comment = UNSET
        else:
            parent_comment = WorkitemCommentsListGetResponseDataItemRelationshipsParentComment.from_dict(
                _parent_comment
            )

        _project = d.pop("project", UNSET)
        project: Union[
            Unset, WorkitemCommentsListGetResponseDataItemRelationshipsProject
        ]
        if isinstance(_project, Unset):
            project = UNSET
        else:
            project = WorkitemCommentsListGetResponseDataItemRelationshipsProject.from_dict(
                _project
            )

        workitem_comments_list_get_response_data_item_relationships_obj = cls(
            author=author,
            child_comments=child_comments,
            parent_comment=parent_comment,
            project=project,
        )

        workitem_comments_list_get_response_data_item_relationships_obj.additional_properties = (
            d
        )
        return workitem_comments_list_get_response_data_item_relationships_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="LinkedoslcresourcesListPostResponseDataItemLinks")


@_attrs_define
class LinkedoslcresourcesListPostResponseDataItemLinks:
    """"""

    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        linkedoslcresources_list_post_response_data_item_links_obj = cls()

        linkedoslcresources_list_post_response_data_item_links_obj.additional_properties = (
            d
        )
        return linkedoslcresources_list_post_response_data_item_links_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Data model classes returned by the client."""
from __future__ import annotations

import dataclasses
import datetime
import enum
import typing as t
import warnings

__all__ = [
    "Document",
    "DocumentReference",
    "HtmlContent",
    "Layouter",
    "RenderingLayout",
    "RenderingProperties",
    "SelectTestCasesBy",
    "StatusItem",
    "TestRecord",
    "TestRun",
    "TextContent",
    "WorkItem",
    "WorkItemAttachment",
    "WorkItemLink",
]

BOOLEAN_RENDERING_PROPERTIES = ["fieldsAtEndAsTable", "hidden"]
RENDERING_LAYOUT_FIELDS = {
    "fieldsAtEndAsTable": "fields_at_end_as_table",
    "fieldsAtStart": "fields_at_start",
    "fieldsAtEnd": "fields_at_end",
    "sidebarWorkItemFields": "sidebar_work_item_fields",
    "hidden": "hidden",
}


@dataclasses.dataclass
class StatusItem:
    """A parent data class for WorkItem and Document."""

    id: str | None = None
    type: str | None = None
    status: str | None = None

    def to_dict(self) -> dict[str, t.Any]:
        """Return the content of the StatusItem as dictionary."""
        return {
            "id": self.id,
            "type": self.type,
            "status": self.status,
        }


@dataclasses.dataclass
class DocumentReference:
    """A reference to a document to be used in relations."""

    module_folder: str
    module_name: str


class WorkItem(StatusItem):
    """A data class containing all relevant data of a Polarion WorkItem."""

    title: str | None = None
    description: TextContent | None = None
    additional_attributes: dict[str, t.Any] = {}
    linked_work_items: list[WorkItemLink] = []
    attachments: list[WorkItemAttachment] = []
    linked_work_items_truncated: bool = False
    attachments_truncated: bool = False
    home_document: DocumentReference | None = None

    def __init__(
        self,
        id: str | None = None,
        *,
        title: str | None = None,
        description_type: str | None = None,
        description: TextContent | str | None = None,
        type: str | None = None,
        status: str | None = None,
        additional_attributes: dict[str, t.Any] | None = None,
        linked_work_items: list[WorkItemLink] | None = None,
        attachments: list[WorkItemAttachment] | None = None,
        linked_work_items_truncated: bool = False,
        attachments_truncated: bool = False,
        home_document: DocumentReference | None = None,
        **kwargs,
    ):
        super().__init__(id, type, status)
        self.title = title
        if description_type or isinstance(description, str):
            warnings.warn(
                "Using description as str or description_type is "
                "deprecated. Use TextContent instead.",
                DeprecationWarning,
                stacklevel=2,
            )

            assert description_type, (
                "You have to set a description_type when using a string as"
                " description"
            )
            assert not isinstance(description, TextContent), (
                "Don't use description_type when setting description as "
                "TextContent"
            )
            description = TextContent(description_type, description)
        self.description = description
        self.additional_attributes = (additional_attributes or {}) | kwargs
        self.linked_work_items = linked_work_items or []
        self.attachments = attachments or []
        self.linked_work_items_truncated = linked_work_items_truncated
        self.attachments_truncated = attachments_truncated
        self.home_document = home_document

    def __getattribute__(self, item: str) -> t.Any:
        """Return all non WorkItem attributes from additional_properties."""
        if item.startswith("__") or item in dir(self.__class__):
            return super().__getattribute__(item)
        return self.additional_attributes.get(item)

    def __setattr__(self, key: str, value: t.Any):
        """Set all non WorkItem attributes in additional_properties."""
        if key in dir(self.__class__):
            super().__setattr__(key, value)
        else:
            self.additional_attributes[key] = value

    def to_dict(self) -> dict[str, t.Any]:
        """Return the content of the WorkItem as dictionary."""
        sorted_links = sorted(
            self.linked_work_items,
            key=lambda x: f"{x.role}/{x.secondary_work_item_project}/{x.secondary_work_item_id}",  # pylint: disable=line-too-long
        )

        sorted_attachments = sorted(
            self.attachments, key=lambda x: x.file_name or ""
        )

        return {
            "id": self.id,
            "title": self.title,
            "description": self.description,
            "type": self.type,
            "status": self.status,
            "additional_attributes": dict(
                sorted(self.additional_attributes.items())
            ),
            "linked_work_items": [
                dataclasses.asdict(lwi) for lwi in sorted_links
            ],
            "attachments": [
                dataclasses.asdict(at) for at in sorted_attachments
            ],
            "home_document": (
                dataclasses.asdict(self.home_document)
                if self.home_document
                else None
            ),
        }


@dataclasses.dataclass
class WorkItemLink:
    """A link between multiple Polarion WorkItems.

    The primary_work_item_id is the ID of the owner of the link, the
    secondary_work_item_id represents the linked workitem.
    """

    primary_work_item_id: str
    secondary_work_item_id: str
    role: str
    suspect: bool | None = None
    secondary_work_item_project: str | None = (
        None  # Use to set differing project
    )


@dataclasses.dataclass
class WorkItemAttachment:
    """An Attachment of a WorkItem."""

    work_item_id: str
    id: str
    title: str | None = None
    content_bytes: bytes | None = None
    mime_type: str | None = None
    file_name: str | None = None


class Document(StatusItem):
    """A data class containing all relevant data of a Polarion Document."""

    module_folder: str | None = None
    module_name: str | None = None
    home_page_content: TextContent | None = None
    title: str | None = None
    outline_numbering: bool | None = None
    outline_numbering_prefix: str | None = None

    def __init__(
        self,
        id: str | None = None,
        module_folder: str | None = None,
        module_name: str | None = None,
        type: str | None = None,
        status: str | None = None,
        home_page_content: TextContent | None = None,
        title: str | None = None,
        rendering_layouts: list[RenderingLayout] | None = None,
        outline_numbering: bool | None = None,
        outline_numbering_prefix: str | None = None,
    ):
        super().__init__(id, type, status)
        self.module_folder = module_folder
        self.module_name = module_name
        self.home_page_content = home_page_content
        self.title = title
        self.rendering_layouts = rendering_layouts
        self.outline_numbering = outline_numbering
        self.outline_numbering_prefix = outline_numbering_prefix

    def __eq__(self, other):
        """Compare dicts instead of hashes."""
        return self.__dict__ == other.__dict__


@dataclasses.dataclass
class RenderingLayout:
    """A class to describe how a work item should be rendered in a document."""

    label: str | None = None
    layouter: Layouter | None = None
    properties: RenderingProperties | None = None
    type: str | None = None

    def __init__(
        self,
        label: str | None = None,
        layouter: Layouter | str | None = None,
        properties: list[dict[str, t.Any]] | RenderingProperties | None = None,
        type: str | None = None,
    ):
        if isinstance(layouter, str):
            layouter = Layouter(layouter)

        if isinstance(properties, list):
            _properties: list[dict[str, t.Any]] = properties
            properties = RenderingProperties()
            for prop in _properties:
                key = prop["key"]
                value = prop.get("value", "")
                if key in BOOLEAN_RENDERING_PROPERTIES:
                    value = value == "true"
                else:
                    value = value.split(",")
                setattr(properties, RENDERING_LAYOUT_FIELDS[key], value)

        self.label = label
        self.layouter = layouter
        self.properties = properties
        self.type = type


class Layouter(enum.Enum):
    """Layout selection for work items in documents."""

    DEFAULT = "default"  # Seems to be title only
    TITLE = "title"  # Title only
    PARAGRAPH = "paragraph"  # Description only
    SECTION = "section"  # Title and description
    TITLE_TEST_STEPS = "titleTestSteps"  # Title and testSteps
    TITLE_DESC_TEST_STEPS = "titleDescTestSteps"  # Title,description,testSteps


@dataclasses.dataclass
class RenderingProperties:
    """Properties for custom field rendering of workitems in documents."""

    fields_at_start: list[str] | None = None
    fields_at_end: list[str] | None = None
    sidebar_work_item_fields: list[str] | None = None
    fields_at_end_as_table: bool = False
    hidden: bool = False

    def serialize(self) -> list[dict[str, t.Any]]:
        """Serialize an instance of this class to be sent via the API."""
        result = []
        for pol_key, key in RENDERING_LAYOUT_FIELDS.items():
            if (value := getattr(self, key)) is not None:
                if isinstance(value, list):
                    value = ",".join(value)
                else:
                    value = str(value).lower()
                result.append({"key": pol_key, "value": value})
        return result


@dataclasses.dataclass
class TestRun(StatusItem):
    """A data class for all data of a test run."""

    title: str | None = None
    home_page_content: TextContent | None = None
    finished_on: datetime.datetime | None = None
    group_id: str | None = None
    id_prefix: str | None = None
    is_template: bool | None = None
    keep_in_history: bool | None = None
    query: str | None = None
    use_report_from_template: bool | None = None
    select_test_cases_by: SelectTestCasesBy | None = None
    additional_attributes: dict[str, t.Any] = dataclasses.field(
        default_factory=dict
    )


@dataclasses.dataclass
class TestRecord:
    """A data class for test record data."""

    test_run_id: str
    work_item_project_id: str
    work_item_id: str
    work_item_revision: str | None = None
    iteration: int = -1
    duration: float = -1
    result: str | None = None
    comment: TextContent | None = None
    executed: datetime.datetime | None = None
    additional_attributes: dict[str, t.Any] = dataclasses.field(
        default_factory=dict
    )


class TextContent(dict):
    """A data class for text content in Polarion."""

    def __init__(self, type: str | None, value: str | None):
        super().__init__(type=type, value=value)

    @property
    def type(self):
        """Return type of the TextContent."""
        return self["type"]

    @type.setter
    def type(self, type: str):
        self["type"] = type

    @property
    def value(self):
        """Return value of the TextContent."""
        return self["value"]

    @value.setter
    def value(self, value: str):
        self["value"] = value


class HtmlContent(TextContent):
    """A specialized class for HTML based TextContent."""

    def __init__(self, value: str):
        super().__init__(type="text/html", value=value)


class SelectTestCasesBy(str, enum.Enum):
    """Test case selection mode enum."""

    AUTOMATEDPROCESS = "automatedProcess"
    DYNAMICLIVEDOC = "dynamicLiveDoc"
    DYNAMICQUERYRESULT = "dynamicQueryResult"
    MANUALSELECTION = "manualSelection"
    STATICLIVEDOC = "staticLiveDoc"
    STATICQUERYRESULT = "staticQueryResult"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Base classes for client implementations on project Level."""
import abc
import datetime
import functools
import logging
import random
import time
import typing as t

from polarion_rest_api_client import data_models as dm
from polarion_rest_api_client import errors
from polarion_rest_api_client.open_api_client import models as api_models
from polarion_rest_api_client.open_api_client import types as oa_types

if t.TYPE_CHECKING:
    from polarion_rest_api_client import client as polarion_client

T = t.TypeVar("T")
ST = t.TypeVar("ST", bound=dm.StatusItem)
logger = logging.getLogger(__name__)
_min_sleep = 5
_max_sleep = 15

UT = t.TypeVar("UT", str, int, float, datetime.datetime, bool, None)


class BaseClient(abc.ABC):
    """The overall base client for all project related clients."""

    _retry_methods: list[str] = []

    def __init__(
        self, project_id: str, client: "polarion_client.PolarionClient"
    ):
        self._project_id = project_id
        self._client = client

    def __getattribute__(self, name):
        """Retry method calls defined in _retry_methods."""
        attr = super().__getattribute__(name)
        retry_methods = super().__getattribute__("_retry_methods")
        if name in retry_methods and callable(attr):
            return functools.partial(
                super().__getattribute__("_retry_on_error"), attr
            )

        return attr

    def _handle_text_content(
        self,
        polarion_content: (
            api_models.DocumentsSingleGetResponseDataAttributesHomePageContent  # pylint: disable=line-too-long
            | api_models.TestrecordsListGetResponseDataItemAttributesComment
            | api_models.TestrunsListGetResponseDataItemAttributesHomePageContent
            | oa_types.Unset
        ),
    ) -> dm.TextContent | None:
        if not polarion_content:
            return None

        return dm.TextContent(
            type=str(polarion_content.type) if polarion_content.type else None,
            value=polarion_content.value or None,
        )

    def _build_sparse_fields(
        self, fields_dict: dict[str, str]
    ) -> api_models.SparseFields | oa_types.Unset:
        """Build the SparseFields object based on a dict.

        Ensure that every key follow the pattern 'fields[XXX]'.
        """
        new_field_dict: dict[str, str] = {}
        for key, value in fields_dict.items():
            if key.startswith("fields["):
                new_field_dict[key] = value
            else:
                new_field_dict[f"fields[{key}]"] = value
        return api_models.SparseFields.from_dict(new_field_dict)

    @t.overload
    def unset_to_none(self, value: oa_types.Unset) -> None:
        """Return None if value is Unset, else the value."""

    @t.overload
    def unset_to_none(self, value: UT) -> UT:
        """Return None if value is Unset, else the value."""

    def unset_to_none(self, value: t.Any) -> t.Any:
        """Return None if value is Unset, else the value."""
        if isinstance(value, oa_types.Unset):
            return None
        return value

    def _raise_on_error(self, response: oa_types.Response):
        def unexpected_error():
            return errors.PolarionApiUnexpectedException(
                response.status_code, response.content
            )

        if response.status_code not in range(400, 600):
            return

        if (
            isinstance(response.parsed, api_models.Errors)
            and response.parsed.errors
        ):
            raise errors.PolarionApiException(
                response.status_code,
                *[
                    (
                        e.status,
                        e.detail,
                        (
                            e.source.pointer
                            if not (
                                isinstance(e.source, oa_types.Unset)
                                or e.source is None
                            )
                            else "No error pointer"
                        ),
                    )
                    for e in response.parsed.errors
                ],
            )
        raise unexpected_error()

    def _retry_on_error(self, call: t.Callable, *args: t.Any, **kwargs: t.Any):
        try:
            return call(*args, **kwargs)
        except Exception as e:
            if isinstance(e, errors.PolarionApiException):
                if e.args[0] == 404:
                    raise e
            logger.warning(
                "Will retry after failing on first attempt, "
                "due to the following error %s",
                e,
            )
            time.sleep(random.uniform(_min_sleep, _max_sleep))
            return call(*args, **kwargs)


class ItemsClient(BaseClient, t.Generic[T], abc.ABC):
    """A client for items of a project, which can be created or requested."""

    _retry_methods = ["get_multi", "get", "_create", "_delete"]

    @abc.abstractmethod
    def get_multi(
        self,
        *args: t.Any,
        page_size: int = 100,
        page_number: int = 1,
        **kwargs: t.Any,
    ) -> tuple[list[T], bool]:
        """Get multiple matching items for a specific page.

        In addition, a flag whether a next page is available is
        returned.
        """

    @abc.abstractmethod
    def get(self, *args, **kwargs) -> T | None:
        """Get a specific single item."""
        return self._retry_on_error(self.get, *args, **kwargs)

    def get_all(self, *args, **kwargs) -> list[T]:
        """Return all matching items using get_multi with auto pagination."""
        page = 1
        items, next_page = self.get_multi(
            *args, page_size=self._client.page_size, page_number=page, **kwargs
        )
        while next_page:
            page += 1
            _items, next_page = self.get_multi(
                *args,
                page_size=self._client.page_size,
                page_number=page,
                **kwargs,
            )
            items += _items
        return items

    @abc.abstractmethod
    def _create(self, items: list[T]): ...

    def _split_into_batches(
        self, items: list[T]
    ) -> t.Generator[list[T], None, None]:
        for i in range(0, len(items), self._client.batch_size):
            yield items[i : i + self._client.batch_size]

    def create(self, items: T | list[T]):
        """Create one or multiple items."""
        if not isinstance(items, list):
            items = [items]

        for batch in self._split_into_batches(items):
            self._create(batch)

    @abc.abstractmethod
    def _delete(self, items: list[T]): ...

    def delete(self, items: T | list[T]):
        """Delete one or multiple items."""
        if not isinstance(items, list):
            items = [items]
        for batch in self._split_into_batches(items):
            self._delete(batch)


class UpdatableItemsClient(ItemsClient, t.Generic[T], abc.ABC):
    """A client for items which can also be updated."""

    _retry_methods = ["get_multi", "get", "_create", "_delete", "_update"]

    def _split_into_update_batches(
        self, items: list[T]
    ) -> t.Generator[list[T], None, None] | t.Generator[T, None, None]:
        yield from self._split_into_batches(items)

    @abc.abstractmethod
    def _update(self, to_update: T | list[T]): ...

    def update(self, items: T | list[T]):
        """Update the provided item or items."""
        if not isinstance(items, list):
            items = [items]

        for batch in self._split_into_update_batches(items):
            self._update(batch)


class SingleUpdatableItemsMixin(t.Generic[T]):
    """Mixin to split batches into single items."""

    def _split_into_update_batches(
        self, items: list[T]
    ) -> t.Generator[T, None, None]:
        yield from items


class StatusItemClient(UpdatableItemsClient, t.Generic[ST], abc.ABC):
    """A client for items, which have a status.

    We support to set a specific status for these instead of deleting
    them. This status has to be provided on initialization.
    """

    def __init__(
        self,
        project_id: str,
        client: "polarion_client.PolarionClient",
        delete_status: str | None = None,
    ):
        super().__init__(project_id, client)
        self.delete_status = delete_status

    def delete(self, items: ST | list[ST]):
        """Delete the item if no delete_status was set, else update status."""
        if self.delete_status is None:
            super().delete(items)
        else:
            if not isinstance(items, list):
                items = [items]
            for item in items:
                item.status = self.delete_status
            self.update(items)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.documents_single_post_response_data_attributes_rendering_layouts_item_properties_item import (
        DocumentsSinglePostResponseDataAttributesRenderingLayoutsItemPropertiesItem,
    )


T = TypeVar(
    "T", bound="DocumentsSinglePostResponseDataAttributesRenderingLayoutsItem"
)


@_attrs_define
class DocumentsSinglePostResponseDataAttributesRenderingLayoutsItem:
    """
    Attributes:
        label (Union[Unset, str]):  Example: My label.
        layouter (Union[Unset, str]):  Example: paragraph.
        properties (Union[Unset, List['DocumentsSinglePostResponseDataAttributesRenderingLayoutsItemPropertiesItem']]):
        type (Union[Unset, str]):  Example: task.
    """

    label: Union[Unset, str] = UNSET
    layouter: Union[Unset, str] = UNSET
    properties: Union[
        Unset,
        List[
            "DocumentsSinglePostResponseDataAttributesRenderingLayoutsItemPropertiesItem"
        ],
    ] = UNSET
    type: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        label = self.label

        layouter = self.layouter

        properties: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.properties, Unset):
            properties = []
            for properties_item_data in self.properties:
                properties_item = properties_item_data.to_dict()
                properties.append(properties_item)

        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if label is not UNSET:
            field_dict["label"] = label
        if layouter is not UNSET:
            field_dict["layouter"] = layouter
        if properties is not UNSET:
            field_dict["properties"] = properties
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.documents_single_post_response_data_attributes_rendering_layouts_item_properties_item import (
            DocumentsSinglePostResponseDataAttributesRenderingLayoutsItemPropertiesItem,
        )

        d = src_dict.copy()
        label = d.pop("label", UNSET)

        layouter = d.pop("layouter", UNSET)

        properties = []
        _properties = d.pop("properties", UNSET)
        for properties_item_data in _properties or []:
            properties_item = DocumentsSinglePostResponseDataAttributesRenderingLayoutsItemPropertiesItem.from_dict(
                properties_item_data
            )

            properties.append(properties_item)

        type = d.pop("type", UNSET)

        documents_single_post_response_data_attributes_rendering_layouts_item_obj = cls(
            label=label,
            layouter=layouter,
            properties=properties,
            type=type,
        )

        documents_single_post_response_data_attributes_rendering_layouts_item_obj.additional_properties = (
            d
        )
        return documents_single_post_response_data_attributes_rendering_layouts_item_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.projecttemplates_list_get_response_data_item_attributes_parameters import (
        ProjecttemplatesListGetResponseDataItemAttributesParameters,
    )


T = TypeVar("T", bound="ProjecttemplatesListGetResponseDataItemAttributes")


@_attrs_define
class ProjecttemplatesListGetResponseDataItemAttributes:
    """
    Attributes:
        custom_icon (Union[Unset, str]):
        description (Union[Unset, str]):
        distributions (Union[Unset, List[str]]):
        id (Union[Unset, str]):
        is_default (Union[Unset, bool]):
        name (Union[Unset, str]):
        parameters (Union[Unset, ProjecttemplatesListGetResponseDataItemAttributesParameters]):
    """

    custom_icon: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    distributions: Union[Unset, List[str]] = UNSET
    id: Union[Unset, str] = UNSET
    is_default: Union[Unset, bool] = UNSET
    name: Union[Unset, str] = UNSET
    parameters: Union[
        Unset, "ProjecttemplatesListGetResponseDataItemAttributesParameters"
    ] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        custom_icon = self.custom_icon

        description = self.description

        distributions: Union[Unset, List[str]] = UNSET
        if not isinstance(self.distributions, Unset):
            distributions = self.distributions

        id = self.id

        is_default = self.is_default

        name = self.name

        parameters: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.parameters, Unset):
            parameters = self.parameters.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if custom_icon is not UNSET:
            field_dict["customIcon"] = custom_icon
        if description is not UNSET:
            field_dict["description"] = description
        if distributions is not UNSET:
            field_dict["distributions"] = distributions
        if id is not UNSET:
            field_dict["id"] = id
        if is_default is not UNSET:
            field_dict["isDefault"] = is_default
        if name is not UNSET:
            field_dict["name"] = name
        if parameters is not UNSET:
            field_dict["parameters"] = parameters

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.projecttemplates_list_get_response_data_item_attributes_parameters import (
            ProjecttemplatesListGetResponseDataItemAttributesParameters,
        )

        d = src_dict.copy()
        custom_icon = d.pop("customIcon", UNSET)

        description = d.pop("description", UNSET)

        distributions = cast(List[str], d.pop("distributions", UNSET))

        id = d.pop("id", UNSET)

        is_default = d.pop("isDefault", UNSET)

        name = d.pop("name", UNSET)

        _parameters = d.pop("parameters", UNSET)
        parameters: Union[
            Unset, ProjecttemplatesListGetResponseDataItemAttributesParameters
        ]
        if isinstance(_parameters, Unset):
            parameters = UNSET
        else:
            parameters = ProjecttemplatesListGetResponseDataItemAttributesParameters.from_dict(
                _parameters
            )

        projecttemplates_list_get_response_data_item_attributes_obj = cls(
            custom_icon=custom_icon,
            description=description,
            distributions=distributions,
            id=id,
            is_default=is_default,
            name=name,
            parameters=parameters,
        )

        projecttemplates_list_get_response_data_item_attributes_obj.additional_properties = (
            d
        )
        return projecttemplates_list_get_response_data_item_attributes_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.errors import Errors
from ...models.workitem_approvals_list_delete_request import (
    WorkitemApprovalsListDeleteRequest,
)
from ...types import Response


def _get_kwargs(
    project_id: str,
    work_item_id: str,
    *,
    body: WorkitemApprovalsListDeleteRequest,
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}

    _kwargs: Dict[str, Any] = {
        "method": "delete",
        "url": "/projects/{projectId}/workitems/{workItemId}/approvals".format(
            projectId=project_id,
            workItemId=work_item_id,
        ),
    }

    _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[Any, Errors]]:
    if response.status_code == HTTPStatus.NO_CONTENT:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = Errors.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = Errors.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = Errors.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = Errors.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.CONFLICT:
        response_409 = Errors.from_dict(response.json())

        return response_409
    if response.status_code == HTTPStatus.REQUEST_ENTITY_TOO_LARGE:
        response_413 = Errors.from_dict(response.json())

        return response_413
    if response.status_code == HTTPStatus.UNSUPPORTED_MEDIA_TYPE:
        response_415 = Errors.from_dict(response.json())

        return response_415
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = Errors.from_dict(response.json())

        return response_500
    if response.status_code == HTTPStatus.SERVICE_UNAVAILABLE:
        response_503 = Errors.from_dict(response.json())

        return response_503
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, Errors]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    project_id: str,
    work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: WorkitemApprovalsListDeleteRequest,
) -> Response[Union[Any, Errors]]:
    """Deletes a list of Work Item Approvals.

    Args:
        project_id (str):
        work_item_id (str):
        body (WorkitemApprovalsListDeleteRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, Errors]]
    """

    kwargs = _get_kwargs(
        project_id=project_id,
        work_item_id=work_item_id,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    project_id: str,
    work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: WorkitemApprovalsListDeleteRequest,
) -> Optional[Union[Any, Errors]]:
    """Deletes a list of Work Item Approvals.

    Args:
        project_id (str):
        work_item_id (str):
        body (WorkitemApprovalsListDeleteRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, Errors]
    """

    return sync_detailed(
        project_id=project_id,
        work_item_id=work_item_id,
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    project_id: str,
    work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: WorkitemApprovalsListDeleteRequest,
) -> Response[Union[Any, Errors]]:
    """Deletes a list of Work Item Approvals.

    Args:
        project_id (str):
        work_item_id (str):
        body (WorkitemApprovalsListDeleteRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, Errors]]
    """

    kwargs = _get_kwargs(
        project_id=project_id,
        work_item_id=work_item_id,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    project_id: str,
    work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: WorkitemApprovalsListDeleteRequest,
) -> Optional[Union[Any, Errors]]:
    """Deletes a list of Work Item Approvals.

    Args:
        project_id (str):
        work_item_id (str):
        body (WorkitemApprovalsListDeleteRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, Errors]
    """

    return (
        await asyncio_detailed(
            project_id=project_id,
            work_item_id=work_item_id,
            client=client,
            body=body,
        )
    ).parsed



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="BranchDocumentsRequestBodyDocumentConfigurationsItem")


@_attrs_define
class BranchDocumentsRequestBodyDocumentConfigurationsItem:
    """
    Attributes:
        source_document (str): Reference path of the source Document. Example: MyProjectId/MySpaceId/MyDocumentId.
        copy_workflow_status_and_signatures (Union[Unset, bool]): Specifies that workflow status and signatures should
            be copied to the branched document.
        initialized_fields (Union[Unset, List[str]]): Specifies fields of overwritten Work Items that should be
            initialized (instead of being copied from source Work Items).
        overwrite_work_items (Union[Unset, bool]): Specifies that Work Items in the branched Document should be
            overwritten (instead of being referenced).
        query (Union[Unset, str]): Specifies optional filtering query. Example: status:open.
        source_revision (Union[Unset, str]): Revision of the source Document. Example: 1234.
        target_document_name (Union[Unset, str]): Name for new Document. Example: MyDocumentId.
        target_document_title (Union[Unset, str]): Title for new Document. Example: My Document Title.
        target_project_id (Union[Unset, str]): Project where new document will be created. Example: MyProjectId.
        target_space_id (Union[Unset, str]): Space where new document will be created. Example: MySpaceId.
        update_title_heading (Union[Unset, bool]): Specifies that title heading of the target Document should be set to
            the new Document's title.
    """

    source_document: str
    copy_workflow_status_and_signatures: Union[Unset, bool] = UNSET
    initialized_fields: Union[Unset, List[str]] = UNSET
    overwrite_work_items: Union[Unset, bool] = UNSET
    query: Union[Unset, str] = UNSET
    source_revision: Union[Unset, str] = UNSET
    target_document_name: Union[Unset, str] = UNSET
    target_document_title: Union[Unset, str] = UNSET
    target_project_id: Union[Unset, str] = UNSET
    target_space_id: Union[Unset, str] = UNSET
    update_title_heading: Union[Unset, bool] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        source_document = self.source_document

        copy_workflow_status_and_signatures = (
            self.copy_workflow_status_and_signatures
        )

        initialized_fields: Union[Unset, List[str]] = UNSET
        if not isinstance(self.initialized_fields, Unset):
            initialized_fields = self.initialized_fields

        overwrite_work_items = self.overwrite_work_items

        query = self.query

        source_revision = self.source_revision

        target_document_name = self.target_document_name

        target_document_title = self.target_document_title

        target_project_id = self.target_project_id

        target_space_id = self.target_space_id

        update_title_heading = self.update_title_heading

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "sourceDocument": source_document,
            }
        )
        if copy_workflow_status_and_signatures is not UNSET:
            field_dict["copyWorkflowStatusAndSignatures"] = (
                copy_workflow_status_and_signatures
            )
        if initialized_fields is not UNSET:
            field_dict["initializedFields"] = initialized_fields
        if overwrite_work_items is not UNSET:
            field_dict["overwriteWorkItems"] = overwrite_work_items
        if query is not UNSET:
            field_dict["query"] = query
        if source_revision is not UNSET:
            field_dict["sourceRevision"] = source_revision
        if target_document_name is not UNSET:
            field_dict["targetDocumentName"] = target_document_name
        if target_document_title is not UNSET:
            field_dict["targetDocumentTitle"] = target_document_title
        if target_project_id is not UNSET:
            field_dict["targetProjectId"] = target_project_id
        if target_space_id is not UNSET:
            field_dict["targetSpaceId"] = target_space_id
        if update_title_heading is not UNSET:
            field_dict["updateTitleHeading"] = update_title_heading

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        source_document = d.pop("sourceDocument")

        copy_workflow_status_and_signatures = d.pop(
            "copyWorkflowStatusAndSignatures", UNSET
        )

        initialized_fields = cast(List[str], d.pop("initializedFields", UNSET))

        overwrite_work_items = d.pop("overwriteWorkItems", UNSET)

        query = d.pop("query", UNSET)

        source_revision = d.pop("sourceRevision", UNSET)

        target_document_name = d.pop("targetDocumentName", UNSET)

        target_document_title = d.pop("targetDocumentTitle", UNSET)

        target_project_id = d.pop("targetProjectId", UNSET)

        target_space_id = d.pop("targetSpaceId", UNSET)

        update_title_heading = d.pop("updateTitleHeading", UNSET)

        branch_documents_request_body_document_configurations_item_obj = cls(
            source_document=source_document,
            copy_workflow_status_and_signatures=copy_workflow_status_and_signatures,
            initialized_fields=initialized_fields,
            overwrite_work_items=overwrite_work_items,
            query=query,
            source_revision=source_revision,
            target_document_name=target_document_name,
            target_document_title=target_document_title,
            target_project_id=target_project_id,
            target_space_id=target_space_id,
            update_title_heading=update_title_heading,
        )

        branch_documents_request_body_document_configurations_item_obj.additional_properties = (
            d
        )
        return branch_documents_request_body_document_configurations_item_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.teststep_results_list_post_request_data_item_attributes_comment import (
        TeststepResultsListPostRequestDataItemAttributesComment,
    )


T = TypeVar("T", bound="TeststepResultsListPostRequestDataItemAttributes")


@_attrs_define
class TeststepResultsListPostRequestDataItemAttributes:
    """
    Attributes:
        comment (Union[Unset, TeststepResultsListPostRequestDataItemAttributesComment]):
        result (Union[Unset, str]):  Example: passed.
    """

    comment: Union[
        Unset, "TeststepResultsListPostRequestDataItemAttributesComment"
    ] = UNSET
    result: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        comment: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.comment, Unset):
            comment = self.comment.to_dict()

        result = self.result

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if comment is not UNSET:
            field_dict["comment"] = comment
        if result is not UNSET:
            field_dict["result"] = result

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.teststep_results_list_post_request_data_item_attributes_comment import (
            TeststepResultsListPostRequestDataItemAttributesComment,
        )

        d = src_dict.copy()
        _comment = d.pop("comment", UNSET)
        comment: Union[
            Unset, TeststepResultsListPostRequestDataItemAttributesComment
        ]
        if isinstance(_comment, Unset):
            comment = UNSET
        else:
            comment = TeststepResultsListPostRequestDataItemAttributesComment.from_dict(
                _comment
            )

        result = d.pop("result", UNSET)

        teststep_results_list_post_request_data_item_attributes_obj = cls(
            comment=comment,
            result=result,
        )

        teststep_results_list_post_request_data_item_attributes_obj.additional_properties = (
            d
        )
        return teststep_results_list_post_request_data_item_attributes_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.testruns_single_patch_request_data_relationships_document import (
        TestrunsSinglePatchRequestDataRelationshipsDocument,
    )
    from ..models.testruns_single_patch_request_data_relationships_project_span import (
        TestrunsSinglePatchRequestDataRelationshipsProjectSpan,
    )
    from ..models.testruns_single_patch_request_data_relationships_summary_defect import (
        TestrunsSinglePatchRequestDataRelationshipsSummaryDefect,
    )


T = TypeVar("T", bound="TestrunsSinglePatchRequestDataRelationships")


@_attrs_define
class TestrunsSinglePatchRequestDataRelationships:
    """
    Attributes:
        document (Union[Unset, TestrunsSinglePatchRequestDataRelationshipsDocument]):
        project_span (Union[Unset, TestrunsSinglePatchRequestDataRelationshipsProjectSpan]):
        summary_defect (Union[Unset, TestrunsSinglePatchRequestDataRelationshipsSummaryDefect]):
    """

    document: Union[
        Unset, "TestrunsSinglePatchRequestDataRelationshipsDocument"
    ] = UNSET
    project_span: Union[
        Unset, "TestrunsSinglePatchRequestDataRelationshipsProjectSpan"
    ] = UNSET
    summary_defect: Union[
        Unset, "TestrunsSinglePatchRequestDataRelationshipsSummaryDefect"
    ] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        document: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.document, Unset):
            document = self.document.to_dict()

        project_span: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.project_span, Unset):
            project_span = self.project_span.to_dict()

        summary_defect: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.summary_defect, Unset):
            summary_defect = self.summary_defect.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if document is not UNSET:
            field_dict["document"] = document
        if project_span is not UNSET:
            field_dict["projectSpan"] = project_span
        if summary_defect is not UNSET:
            field_dict["summaryDefect"] = summary_defect

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.testruns_single_patch_request_data_relationships_document import (
            TestrunsSinglePatchRequestDataRelationshipsDocument,
        )
        from ..models.testruns_single_patch_request_data_relationships_project_span import (
            TestrunsSinglePatchRequestDataRelationshipsProjectSpan,
        )
        from ..models.testruns_single_patch_request_data_relationships_summary_defect import (
            TestrunsSinglePatchRequestDataRelationshipsSummaryDefect,
        )

        d = src_dict.copy()
        _document = d.pop("document", UNSET)
        document: Union[
            Unset, TestrunsSinglePatchRequestDataRelationshipsDocument
        ]
        if isinstance(_document, Unset):
            document = UNSET
        else:
            document = (
                TestrunsSinglePatchRequestDataRelationshipsDocument.from_dict(
                    _document
                )
            )

        _project_span = d.pop("projectSpan", UNSET)
        project_span: Union[
            Unset, TestrunsSinglePatchRequestDataRelationshipsProjectSpan
        ]
        if isinstance(_project_span, Unset):
            project_span = UNSET
        else:
            project_span = TestrunsSinglePatchRequestDataRelationshipsProjectSpan.from_dict(
                _project_span
            )

        _summary_defect = d.pop("summaryDefect", UNSET)
        summary_defect: Union[
            Unset, TestrunsSinglePatchRequestDataRelationshipsSummaryDefect
        ]
        if isinstance(_summary_defect, Unset):
            summary_defect = UNSET
        else:
            summary_defect = TestrunsSinglePatchRequestDataRelationshipsSummaryDefect.from_dict(
                _summary_defect
            )

        testruns_single_patch_request_data_relationships_obj = cls(
            document=document,
            project_span=project_span,
            summary_defect=summary_defect,
        )

        testruns_single_patch_request_data_relationships_obj.additional_properties = (
            d
        )
        return testruns_single_patch_request_data_relationships_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class TestrecordAttachmentsListPostResponseDataItemType(str, Enum):
    TESTRECORD_ATTACHMENTS = "testrecord_attachments"

    def __str__(self) -> str:
        return str(self.value)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Script to fix the specification and build code from it.

Usage: needs 2 args for execution. First one is either 'url' or 'path',
second one is the path to the Open API Spec. E.g.
./build_client_source.sh path /download/spec.json will take the spec
from the given path
"""
import json
import os
import pathlib
import re
import shutil
import subprocess
import sys
import tempfile

import httpx

error_code_pattern = re.compile("[4,5][0-9]{2}")
script_path = pathlib.Path(os.path.dirname(os.path.realpath(__file__)))
rest_api_path = script_path.parent / "polarion_rest_api_client"
template_path = script_path / "custom_templates"
config_path = script_path / "config.yaml"
autoflake_commands = [
    "autoflake",
    "-i",
    "-r",
    "--remove-all-unused-imports",
    "--remove-unused-variables",
    "--ignore-init-module-imports",
    "./open_api_client",
]
black_commands = [
    "black",
    "./open_api_client",
]
isort_commands = [
    "isort",
    "./open_api_client",
]


def fix_spec(src: str, path: str | os.PathLike):
    """Fix errors in the specification."""
    if src == "path":
        with open(path, "r") as f:
            spec = json.load(f)
    elif src == "url":
        response = httpx.get(path)
        spec = response.json()
    else:
        raise Exception(
            "you have to provide a file or url keyword as 1st arg."
        )
    spec_paths = spec["paths"]
    if (
        octet_schema := spec_paths.get(
            "/projects/{projectId}/testruns/{testRunId}/actions/importXUnitTestResults",
            {},
        )
        .get("post", {})
        .get("requestBody", {})
        .get("content", {})
        .get("application/octet-stream", {})
        .get("schema")
    ):
        if octet_schema.get("type") == "object":
            octet_schema["type"] = "string"
            octet_schema["format"] = "binary"

    for spec_path in spec_paths.values():
        for operation_description in spec_path.values():
            if responses := operation_description.get("responses"):
                if "4XX-5XX" in responses:
                    for code, resp in responses.items():
                        if error_code_pattern.fullmatch(code):
                            resp["content"] = responses["4XX-5XX"]["content"]
                    del responses["4XX-5XX"]

    schemas = spec["components"]["schemas"]
    if (
        downloads := schemas.get("jobsSingleGetResponse", {})
        .get("properties", {})
        .get("data", {})
        .get("properties", {})
        .get("links", {})
        .get("properties", {})
        .get("downloads")
    ):
        if "items" not in downloads and downloads.get("type") == "array":
            downloads["items"] = {"type": "string"}

    if (
        downloads := schemas.get("jobsSinglePostResponse", {})
        .get("properties", {})
        .get("data", {})
        .get("properties", {})
        .get("links", {})
        .get("properties", {})
        .get("downloads")
    ):
        if "items" not in downloads and downloads.get("type") == "array":
            downloads["items"] = {"type": "string"}

    if (
        error_source := schemas.get("errors", {})
        .get("properties", {})
        .get("errors", {})
        .get("items", {})
        .get("properties", {})
        .get("source")
    ):
        error_source["nullable"] = True
        if resource := error_source.get("properties", {}).get("resource"):
            resource["nullable"] = True

    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        json.dump(spec, f)
        f.close()

        generator_path = shutil.which("openapi-python-client")
        assert generator_path is not None, (
            "Did not find openapi-python-client generator - "
            "please install dev requirements"
        )

        subprocess.run(
            [
                generator_path,
                "update",
                "--meta",
                "none",
                "--path",
                f.name,
                f"--custom-template-path={template_path}",
                "--config",
                config_path,
            ],
            cwd=rest_api_path,
            check=True,
        )

    subprocess.run(autoflake_commands, cwd=rest_api_path, check=True)
    subprocess.run(black_commands, cwd=rest_api_path, check=True)
    subprocess.run(isort_commands, cwd=rest_api_path, check=True)


if __name__ == "__main__":
    fix_spec(sys.argv[1], sys.argv[2])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class DocumentPartsSingleGetResponseDataRelationshipsNextPartDataType(
    str, Enum
):
    DOCUMENT_PARTS = "document_parts"

    def __str__(self) -> str:
        return str(self.value)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Pagination")


@_attrs_define
class Pagination:
    """
    Attributes:
        calculated_offset (Union[Unset, int]):
        page_number (Union[Unset, int]):
        page_size (Union[Unset, int]):
    """

    calculated_offset: Union[Unset, int] = UNSET
    page_number: Union[Unset, int] = UNSET
    page_size: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        calculated_offset = self.calculated_offset

        page_number = self.page_number

        page_size = self.page_size

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if calculated_offset is not UNSET:
            field_dict["calculatedOffset"] = calculated_offset
        if page_number is not UNSET:
            field_dict["pageNumber"] = page_number
        if page_size is not UNSET:
            field_dict["pageSize"] = page_size

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        calculated_offset = d.pop("calculatedOffset", UNSET)

        page_number = d.pop("pageNumber", UNSET)

        page_size = d.pop("pageSize", UNSET)

        pagination_obj = cls(
            calculated_offset=calculated_offset,
            page_number=page_number,
            page_size=page_size,
        )

        pagination_obj.additional_properties = d
        return pagination_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.testrecords_single_patch_request_data_relationships_defect import (
        TestrecordsSinglePatchRequestDataRelationshipsDefect,
    )
    from ..models.testrecords_single_patch_request_data_relationships_executed_by import (
        TestrecordsSinglePatchRequestDataRelationshipsExecutedBy,
    )


T = TypeVar("T", bound="TestrecordsSinglePatchRequestDataRelationships")


@_attrs_define
class TestrecordsSinglePatchRequestDataRelationships:
    """
    Attributes:
        defect (Union[Unset, TestrecordsSinglePatchRequestDataRelationshipsDefect]):
        executed_by (Union[Unset, TestrecordsSinglePatchRequestDataRelationshipsExecutedBy]):
    """

    defect: Union[
        Unset, "TestrecordsSinglePatchRequestDataRelationshipsDefect"
    ] = UNSET
    executed_by: Union[
        Unset, "TestrecordsSinglePatchRequestDataRelationshipsExecutedBy"
    ] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        defect: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.defect, Unset):
            defect = self.defect.to_dict()

        executed_by: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.executed_by, Unset):
            executed_by = self.executed_by.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if defect is not UNSET:
            field_dict["defect"] = defect
        if executed_by is not UNSET:
            field_dict["executedBy"] = executed_by

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.testrecords_single_patch_request_data_relationships_defect import (
            TestrecordsSinglePatchRequestDataRelationshipsDefect,
        )
        from ..models.testrecords_single_patch_request_data_relationships_executed_by import (
            TestrecordsSinglePatchRequestDataRelationshipsExecutedBy,
        )

        d = src_dict.copy()
        _defect = d.pop("defect", UNSET)
        defect: Union[
            Unset, TestrecordsSinglePatchRequestDataRelationshipsDefect
        ]
        if isinstance(_defect, Unset):
            defect = UNSET
        else:
            defect = (
                TestrecordsSinglePatchRequestDataRelationshipsDefect.from_dict(
                    _defect
                )
            )

        _executed_by = d.pop("executedBy", UNSET)
        executed_by: Union[
            Unset, TestrecordsSinglePatchRequestDataRelationshipsExecutedBy
        ]
        if isinstance(_executed_by, Unset):
            executed_by = UNSET
        else:
            executed_by = TestrecordsSinglePatchRequestDataRelationshipsExecutedBy.from_dict(
                _executed_by
            )

        testrecords_single_patch_request_data_relationships_obj = cls(
            defect=defect,
            executed_by=executed_by,
        )

        testrecords_single_patch_request_data_relationships_obj.additional_properties = (
            d
        )
        return testrecords_single_patch_request_data_relationships_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.errors import Errors
from ...models.linkedworkitems_list_post_request import (
    LinkedworkitemsListPostRequest,
)
from ...models.linkedworkitems_list_post_response import (
    LinkedworkitemsListPostResponse,
)
from ...types import Response


def _get_kwargs(
    project_id: str,
    work_item_id: str,
    *,
    body: LinkedworkitemsListPostRequest,
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}

    _kwargs: Dict[str, Any] = {
        "method": "post",
        "url": "/projects/{projectId}/workitems/{workItemId}/linkedworkitems".format(
            projectId=project_id,
            workItemId=work_item_id,
        ),
    }

    _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[Errors, LinkedworkitemsListPostResponse]]:
    if response.status_code == HTTPStatus.CREATED:
        response_201 = LinkedworkitemsListPostResponse.from_dict(
            response.json()
        )

        return response_201
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = Errors.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = Errors.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = Errors.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = Errors.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.NOT_ACCEPTABLE:
        response_406 = Errors.from_dict(response.json())

        return response_406
    if response.status_code == HTTPStatus.CONFLICT:
        response_409 = Errors.from_dict(response.json())

        return response_409
    if response.status_code == HTTPStatus.REQUEST_ENTITY_TOO_LARGE:
        response_413 = Errors.from_dict(response.json())

        return response_413
    if response.status_code == HTTPStatus.UNSUPPORTED_MEDIA_TYPE:
        response_415 = Errors.from_dict(response.json())

        return response_415
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = Errors.from_dict(response.json())

        return response_500
    if response.status_code == HTTPStatus.SERVICE_UNAVAILABLE:
        response_503 = Errors.from_dict(response.json())

        return response_503
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Errors, LinkedworkitemsListPostResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    project_id: str,
    work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: LinkedworkitemsListPostRequest,
) -> Response[Union[Errors, LinkedworkitemsListPostResponse]]:
    """Creates a list of Linked Work Items.

     Creates the direct outgoing links to other Work Items. (The same as the corresponding Java API
    method.)  Does not pertain to external links or backlinks.

    Args:
        project_id (str):
        work_item_id (str):
        body (LinkedworkitemsListPostRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Errors, LinkedworkitemsListPostResponse]]
    """

    kwargs = _get_kwargs(
        project_id=project_id,
        work_item_id=work_item_id,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    project_id: str,
    work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: LinkedworkitemsListPostRequest,
) -> Optional[Union[Errors, LinkedworkitemsListPostResponse]]:
    """Creates a list of Linked Work Items.

     Creates the direct outgoing links to other Work Items. (The same as the corresponding Java API
    method.)  Does not pertain to external links or backlinks.

    Args:
        project_id (str):
        work_item_id (str):
        body (LinkedworkitemsListPostRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Errors, LinkedworkitemsListPostResponse]
    """

    return sync_detailed(
        project_id=project_id,
        work_item_id=work_item_id,
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    project_id: str,
    work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: LinkedworkitemsListPostRequest,
) -> Response[Union[Errors, LinkedworkitemsListPostResponse]]:
    """Creates a list of Linked Work Items.

     Creates the direct outgoing links to other Work Items. (The same as the corresponding Java API
    method.)  Does not pertain to external links or backlinks.

    Args:
        project_id (str):
        work_item_id (str):
        body (LinkedworkitemsListPostRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Errors, LinkedworkitemsListPostResponse]]
    """

    kwargs = _get_kwargs(
        project_id=project_id,
        work_item_id=work_item_id,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    project_id: str,
    work_item_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    body: LinkedworkitemsListPostRequest,
) -> Optional[Union[Errors, LinkedworkitemsListPostResponse]]:
    """Creates a list of Linked Work Items.

     Creates the direct outgoing links to other Work Items. (The same as the corresponding Java API
    method.)  Does not pertain to external links or backlinks.

    Args:
        project_id (str):
        work_item_id (str):
        body (LinkedworkitemsListPostRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Errors, LinkedworkitemsListPostResponse]
    """

    return (
        await asyncio_detailed(
            project_id=project_id,
            work_item_id=work_item_id,
            client=client,
            body=body,
        )
    ).parsed



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import datetime
from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

T = TypeVar("T", bound="RevisionsSingleGetResponseDataAttributes")


@_attrs_define
class RevisionsSingleGetResponseDataAttributes:
    """
    Attributes:
        created (Union[Unset, datetime.datetime]):  Example: 1970-01-01T00:00:00Z.
        id (Union[Unset, str]):
        internal_commit (Union[Unset, bool]):
        message (Union[Unset, str]):  Example: Message.
        repository_name (Union[Unset, str]):  Example: Repository name.
    """

    created: Union[Unset, datetime.datetime] = UNSET
    id: Union[Unset, str] = UNSET
    internal_commit: Union[Unset, bool] = UNSET
    message: Union[Unset, str] = UNSET
    repository_name: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        created: Union[Unset, str] = UNSET
        if not isinstance(self.created, Unset):
            created = self.created.isoformat()

        id = self.id

        internal_commit = self.internal_commit

        message = self.message

        repository_name = self.repository_name

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if created is not UNSET:
            field_dict["created"] = created
        if id is not UNSET:
            field_dict["id"] = id
        if internal_commit is not UNSET:
            field_dict["internalCommit"] = internal_commit
        if message is not UNSET:
            field_dict["message"] = message
        if repository_name is not UNSET:
            field_dict["repositoryName"] = repository_name

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        _created = d.pop("created", UNSET)
        created: Union[Unset, datetime.datetime]
        if isinstance(_created, Unset):
            created = UNSET
        else:
            created = isoparse(_created)

        id = d.pop("id", UNSET)

        internal_commit = d.pop("internalCommit", UNSET)

        message = d.pop("message", UNSET)

        repository_name = d.pop("repositoryName", UNSET)

        revisions_single_get_response_data_attributes_obj = cls(
            created=created,
            id=id,
            internal_commit=internal_commit,
            message=message,
            repository_name=repository_name,
        )

        revisions_single_get_response_data_attributes_obj.additional_properties = (
            d
        )
        return revisions_single_get_response_data_attributes_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class RelationshipDataBodyType(str, Enum):
    CATEGORIES = "categories"
    DOCUMENTS = "documents"
    DOCUMENT_ATTACHMENTS = "document_attachments"
    DOCUMENT_COMMENTS = "document_comments"
    DOCUMENT_PARTS = "document_parts"
    ENUMERATIONS = "enumerations"
    EXTERNALLYLINKEDWORKITEMS = "externallylinkedworkitems"
    FEATURESELECTIONS = "featureselections"
    GLOBALROLES = "globalroles"
    ICONS = "icons"
    JOBS = "jobs"
    LINKEDOSLCRESOURCES = "linkedoslcresources"
    LINKEDWORKITEMS = "linkedworkitems"
    PAGES = "pages"
    PAGE_ATTACHMENTS = "page_attachments"
    PLANS = "plans"
    PROJECTGROUPS = "projectgroups"
    PROJECTROLES = "projectroles"
    PROJECTS = "projects"
    PROJECTTEMPLATES = "projecttemplates"
    REVISIONS = "revisions"
    SPACES = "spaces"
    TESTPARAMETERS = "testparameters"
    TESTPARAMETER_DEFINITIONS = "testparameter_definitions"
    TESTRECORDS = "testrecords"
    TESTRECORD_ATTACHMENTS = "testrecord_attachments"
    TESTRUNS = "testruns"
    TESTRUN_ATTACHMENTS = "testrun_attachments"
    TESTRUN_COMMENTS = "testrun_comments"
    TESTSTEPRESULT_ATTACHMENTS = "teststepresult_attachments"
    TESTSTEPS = "teststeps"
    TESTSTEP_RESULTS = "teststep_results"
    USERGROUPS = "usergroups"
    USERS = "users"
    WORKITEMS = "workitems"
    WORKITEM_APPROVALS = "workitem_approvals"
    WORKITEM_ATTACHMENTS = "workitem_attachments"
    WORKITEM_COMMENTS = "workitem_comments"
    WORKRECORDS = "workrecords"

    def __str__(self) -> str:
        return str(self.value)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import json

import pytest_httpx

import polarion_rest_api_client as polarion_api
from tests.conftest import TEST_PROJECT_RESPONSE_JSON


def test_api_authentication(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_PROJECT_RESPONSE_JSON, encoding="utf8") as f:
        httpx_mock.add_response(
            match_headers={"Authorization": "Bearer PAT123"},
            json=json.load(f),
        )

    assert client.exists()


def test_check_existing_project(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(TEST_PROJECT_RESPONSE_JSON, encoding="utf8") as f:
        httpx_mock.add_response(json=json.load(f))

    assert client.exists()


def test_check_non_existing_project(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    httpx_mock.add_response(status_code=404, json={})

    assert not client.exists()



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""A client library for accessing Polarion REST API."""

from .client import AuthenticatedClient, Client

__all__ = (
    "AuthenticatedClient",
    "Client",
)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.usergroups_single_get_response_data_meta_errors_item_source import (
        UsergroupsSingleGetResponseDataMetaErrorsItemSource,
    )


T = TypeVar("T", bound="UsergroupsSingleGetResponseDataMetaErrorsItem")


@_attrs_define
class UsergroupsSingleGetResponseDataMetaErrorsItem:
    """
    Attributes:
        detail (Union[Unset, str]): Human-readable explanation specific to this occurrence of the problem. Example:
            Unexpected token, BEGIN_ARRAY expected, but was : BEGIN_OBJECT (at $.data).
        source (Union[Unset, UsergroupsSingleGetResponseDataMetaErrorsItemSource]):
        status (Union[Unset, str]): HTTP status code applicable to this problem. Example: 400.
        title (Union[Unset, str]): Short, human-readable summary of the problem. Example: Bad Request.
    """

    detail: Union[Unset, str] = UNSET
    source: Union[
        Unset, "UsergroupsSingleGetResponseDataMetaErrorsItemSource"
    ] = UNSET
    status: Union[Unset, str] = UNSET
    title: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        detail = self.detail

        source: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.source, Unset):
            source = self.source.to_dict()

        status = self.status

        title = self.title

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if detail is not UNSET:
            field_dict["detail"] = detail
        if source is not UNSET:
            field_dict["source"] = source
        if status is not UNSET:
            field_dict["status"] = status
        if title is not UNSET:
            field_dict["title"] = title

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.usergroups_single_get_response_data_meta_errors_item_source import (
            UsergroupsSingleGetResponseDataMetaErrorsItemSource,
        )

        d = src_dict.copy()
        detail = d.pop("detail", UNSET)

        _source = d.pop("source", UNSET)
        source: Union[
            Unset, UsergroupsSingleGetResponseDataMetaErrorsItemSource
        ]
        if isinstance(_source, Unset):
            source = UNSET
        else:
            source = (
                UsergroupsSingleGetResponseDataMetaErrorsItemSource.from_dict(
                    _source
                )
            )

        status = d.pop("status", UNSET)

        title = d.pop("title", UNSET)

        usergroups_single_get_response_data_meta_errors_item_obj = cls(
            detail=detail,
            source=source,
            status=status,
            title=title,
        )

        usergroups_single_get_response_data_meta_errors_item_obj.additional_properties = (
            d
        )
        return usergroups_single_get_response_data_meta_errors_item_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.testrecords_single_get_response_data_type import (
    TestrecordsSingleGetResponseDataType,
)
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.testrecords_single_get_response_data_attributes import (
        TestrecordsSingleGetResponseDataAttributes,
    )
    from ..models.testrecords_single_get_response_data_links import (
        TestrecordsSingleGetResponseDataLinks,
    )
    from ..models.testrecords_single_get_response_data_meta import (
        TestrecordsSingleGetResponseDataMeta,
    )
    from ..models.testrecords_single_get_response_data_relationships import (
        TestrecordsSingleGetResponseDataRelationships,
    )


T = TypeVar("T", bound="TestrecordsSingleGetResponseData")


@_attrs_define
class TestrecordsSingleGetResponseData:
    """
    Attributes:
        type (Union[Unset, TestrecordsSingleGetResponseDataType]):
        id (Union[Unset, str]):  Example: elibrary/MyTestRunId/MyProjectId/MyTestcaseId/0.
        revision (Union[Unset, str]):  Example: 1234.
        attributes (Union[Unset, TestrecordsSingleGetResponseDataAttributes]):
        relationships (Union[Unset, TestrecordsSingleGetResponseDataRelationships]):
        links (Union[Unset, TestrecordsSingleGetResponseDataLinks]):
        meta (Union[Unset, TestrecordsSingleGetResponseDataMeta]):
    """

    type: Union[Unset, TestrecordsSingleGetResponseDataType] = UNSET
    id: Union[Unset, str] = UNSET
    revision: Union[Unset, str] = UNSET
    attributes: Union[Unset, "TestrecordsSingleGetResponseDataAttributes"] = (
        UNSET
    )
    relationships: Union[
        Unset, "TestrecordsSingleGetResponseDataRelationships"
    ] = UNSET
    links: Union[Unset, "TestrecordsSingleGetResponseDataLinks"] = UNSET
    meta: Union[Unset, "TestrecordsSingleGetResponseDataMeta"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        type: Union[Unset, str] = UNSET
        if not isinstance(self.type, Unset):
            type = self.type.value

        id = self.id

        revision = self.revision

        attributes: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.attributes, Unset):
            attributes = self.attributes.to_dict()

        relationships: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.relationships, Unset):
            relationships = self.relationships.to_dict()

        links: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.links, Unset):
            links = self.links.to_dict()

        meta: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.meta, Unset):
            meta = self.meta.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if type is not UNSET:
            field_dict["type"] = type
        if id is not UNSET:
            field_dict["id"] = id
        if revision is not UNSET:
            field_dict["revision"] = revision
        if attributes is not UNSET:
            field_dict["attributes"] = attributes
        if relationships is not UNSET:
            field_dict["relationships"] = relationships
        if links is not UNSET:
            field_dict["links"] = links
        if meta is not UNSET:
            field_dict["meta"] = meta

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.testrecords_single_get_response_data_attributes import (
            TestrecordsSingleGetResponseDataAttributes,
        )
        from ..models.testrecords_single_get_response_data_links import (
            TestrecordsSingleGetResponseDataLinks,
        )
        from ..models.testrecords_single_get_response_data_meta import (
            TestrecordsSingleGetResponseDataMeta,
        )
        from ..models.testrecords_single_get_response_data_relationships import (
            TestrecordsSingleGetResponseDataRelationships,
        )

        d = src_dict.copy()
        _type = d.pop("type", UNSET)
        type: Union[Unset, TestrecordsSingleGetResponseDataType]
        if isinstance(_type, Unset):
            type = UNSET
        else:
            type = TestrecordsSingleGetResponseDataType(_type)

        id = d.pop("id", UNSET)

        revision = d.pop("revision", UNSET)

        _attributes = d.pop("attributes", UNSET)
        attributes: Union[Unset, TestrecordsSingleGetResponseDataAttributes]
        if isinstance(_attributes, Unset):
            attributes = UNSET
        else:
            attributes = TestrecordsSingleGetResponseDataAttributes.from_dict(
                _attributes
            )

        _relationships = d.pop("relationships", UNSET)
        relationships: Union[
            Unset, TestrecordsSingleGetResponseDataRelationships
        ]
        if isinstance(_relationships, Unset):
            relationships = UNSET
        else:
            relationships = (
                TestrecordsSingleGetResponseDataRelationships.from_dict(
                    _relationships
                )
            )

        _links = d.pop("links", UNSET)
        links: Union[Unset, TestrecordsSingleGetResponseDataLinks]
        if isinstance(_links, Unset):
            links = UNSET
        else:
            links = TestrecordsSingleGetResponseDataLinks.from_dict(_links)

        _meta = d.pop("meta", UNSET)
        meta: Union[Unset, TestrecordsSingleGetResponseDataMeta]
        if isinstance(_meta, Unset):
            meta = UNSET
        else:
            meta = TestrecordsSingleGetResponseDataMeta.from_dict(_meta)

        testrecords_single_get_response_data_obj = cls(
            type=type,
            id=id,
            revision=revision,
            attributes=attributes,
            relationships=relationships,
            links=links,
            meta=meta,
        )

        testrecords_single_get_response_data_obj.additional_properties = d
        return testrecords_single_get_response_data_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Implementation of a client providing work item specific functions."""
import json
import logging
import typing as t

from polarion_rest_api_client import data_models as dm
from polarion_rest_api_client import errors
from polarion_rest_api_client.open_api_client import models as api_models
from polarion_rest_api_client.open_api_client import types as oa_types
from polarion_rest_api_client.open_api_client.api.work_items import (
    delete_work_items,
    get_work_item,
    get_work_items,
    patch_work_item,
    post_work_items,
)

from . import base_classes as bc
from . import work_item_attachments, work_item_links

WT = t.TypeVar("WT", bound=dm.WorkItem)
logger = logging.getLogger(__name__)
if t.TYPE_CHECKING:
    from polarion_rest_api_client import client as polarion_client


def _get_json_content_size(data: dict):
    return len(json.dumps(data).encode("utf-8"))


min_wi_request_size = _get_json_content_size(
    api_models.WorkitemsListPostRequest(data=[]).to_dict()
)


class WorkItems(bc.SingleUpdatableItemsMixin, bc.StatusItemClient):
    """A project specific client for work item operations."""

    def __init__(
        self,
        project_id: str,
        client: "polarion_client.PolarionClient",
        delete_status: str | None = None,
    ):
        super().__init__(project_id, client, delete_status)
        self.attachments = work_item_attachments.WorkItemAttachments(
            project_id, client
        )
        self.links = work_item_links.WorkItemLinks(project_id, client)

    def _update(self, to_update: list[dm.WorkItem] | dm.WorkItem):
        assert not isinstance(to_update, list), "Expected only one item"
        assert to_update.id is not None
        if to_update.type:
            logger.warning(
                "Attempting to change the type of Work Item %s to %s.",
                to_update.id,
                to_update.type,
            )

        response = patch_work_item.sync_detailed(
            self._project_id,
            to_update.id,
            client=self._client.client,
            change_type_to=to_update.type or oa_types.UNSET,
            body=self._build_work_item_patch_request(to_update),
        )
        self._raise_on_error(response)

    @t.overload  # type: ignore[override]
    def get_multi(
        self,
        query: str = "",
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
        work_item_cls: type[WT],
    ) -> tuple[list[WT], bool]:
        """Return the work items on a defined page matching the given query.

        In addition, a flag whether a next page is available is
        returned. Define a fields dictionary as described in the
        Polarion API documentation to get certain fields. This function
        will use the provided WorkItemClass as return type.
        """

    @t.overload
    def get_multi(
        self,
        query: str = "",
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
    ) -> tuple[list[dm.WorkItem], bool]:
        """Return the work items on a defined page matching the given query.

        In addition, a flag whether a next page is available is
        returned. Define a fields dictionary as described in the
        Polarion API documentation to get certain fields.
        """

    def get_multi(
        self,
        query: str = "",
        *,
        page_size: int = 100,
        page_number: int = 1,
        fields: dict[str, str] | None = None,
        work_item_cls=dm.WorkItem,
    ) -> tuple[list[dm.WorkItem], bool] | tuple[list[WT], bool]:
        """Return the work items on a defined page matching the given query.

        In addition, a flag whether a next page is available is
        returned. Define a fields dictionary as described in the
        Polarion API documentation to get certain fields.
        """
        if fields is None:
            fields = self._client.default_fields.workitems

        sparse_fields = self._build_sparse_fields(fields)
        response = get_work_items.sync_detailed(
            self._project_id,
            client=self._client.client,
            fields=sparse_fields,
            query=query,
            pagesize=page_size,
            pagenumber=page_number,
        )

        self._raise_on_error(response)

        work_items_response = response.parsed

        work_items: list[WT] = []

        next_page = False
        if (
            isinstance(
                work_items_response, api_models.WorkitemsListGetResponse
            )
            and work_items_response.data
        ):
            work_items = [
                self._generate_work_item(work_item, work_item_cls)
                for work_item in work_items_response.data
                if not getattr(work_item.meta, "errors", [])
            ]

            next_page = isinstance(
                work_items_response.links,
                api_models.WorkitemsListGetResponseLinks,
            ) and bool(work_items_response.links.next_)

        return work_items, next_page

    @t.overload
    def get(
        self,
        work_item_id: str,
        work_item_cls: type[WT],
        revision: str | None = None,
    ) -> WT | None:
        """Return one specific work item with all fields.

        This also includes all linked work items and attachments. If
        there are to many of these to get them in one request, the
        truncated flags for linked_work_items and attachments will be
        set to True. This function will use the provided WorkItemClass
        as return type.
        """

    @t.overload
    def get(
        self, work_item_id: str, *, revision: str | None = None
    ) -> dm.WorkItem | None:
        """Return one specific work item with all fields.

        This also includes all linked work items and attachments. If
        there are to many of these to get them in one request, the
        truncated flags for linked_work_items and attachments will be
        set to True.
        """

    def get(
        self,
        work_item_id: str,
        work_item_cls=dm.WorkItem,
        revision: str | None = None,
    ) -> WT | dm.WorkItem | None:
        """Return one specific work item with all fields.

        This also includes all linked work items and attachments. If
        there are to many of these to get them in one request, the
        truncated flags for linked_work_items and attachments will be
        set to True.
        """
        response = get_work_item.sync_detailed(
            self._project_id,
            work_item_id,
            client=self._client.client,
            fields=self._build_sparse_fields(
                {
                    "workitems": "@all",
                    "workitem_attachments": "@all",
                    "linkedworkitems": "@all",
                }
            ),
            revision=revision or oa_types.UNSET,
        )
        self._raise_on_error(response)

        if isinstance(
            response.parsed, api_models.WorkitemsSingleGetResponse
        ) and isinstance(
            response.parsed.data, api_models.WorkitemsSingleGetResponseData
        ):
            return self._generate_work_item(
                response.parsed.data, work_item_cls
            )

        return None

    def _create(self, items: list[dm.WorkItem]):
        raise NotImplementedError("We have a custom create instead.")

    def create(self, items: dm.WorkItem | list[dm.WorkItem]):
        """Create WorkItems and respect the max body size of the server."""
        if not isinstance(items, list):
            items = [items]
        current_batch = api_models.WorkitemsListPostRequest(data=[])
        content_size = min_wi_request_size
        batch_start_index = 0
        batch_end_index = 0

        for work_item in items:
            work_item_data = self._build_work_item_post_request(work_item)

            (
                proj_content_size,
                too_big,
            ) = self._calculate_post_work_item_request_sizes(
                work_item_data, content_size
            )

            if too_big:
                raise errors.PolarionWorkItemException(
                    "A WorkItem is too large to create.", work_item
                )

            assert isinstance(current_batch.data, list)
            if (
                proj_content_size >= self._client.max_content_size
                or len(current_batch.data) >= self._client.batch_size
            ):
                self._retry_on_error(
                    self._post_work_item_batch,
                    current_batch,
                    items[batch_start_index:batch_end_index],
                )

                current_batch = api_models.WorkitemsListPostRequest(
                    data=[work_item_data]
                )
                content_size = _get_json_content_size(current_batch.to_dict())
                batch_start_index = batch_end_index
            else:
                assert isinstance(current_batch.data, list)
                current_batch.data.append(work_item_data)
                content_size = proj_content_size

            batch_end_index += 1

        if current_batch.data:
            self._retry_on_error(
                self._post_work_item_batch,
                current_batch,
                items[batch_start_index:],
            )

    def _delete(self, items: list[dm.WorkItem]):
        work_item_ids = [work_item.id for work_item in items]
        response = delete_work_items.sync_detailed(
            self._project_id,
            client=self._client.client,
            body=api_models.WorkitemsListDeleteRequest(
                data=[
                    api_models.WorkitemsListDeleteRequestDataItem(
                        type=api_models.WorkitemsListDeleteRequestDataItemType.WORKITEMS,  # pylint: disable=line-too-long
                        id=f"{self._project_id}/{work_item_id}",
                    )
                    for work_item_id in work_item_ids
                ]
            ),
        )
        self._raise_on_error(response)

    def _build_work_item_post_request(
        self, work_item: dm.WorkItem
    ) -> api_models.WorkitemsListPostRequestDataItem:
        assert work_item.type is not None
        assert work_item.title is not None
        assert work_item.description is not None
        assert work_item.status is not None

        attrs = api_models.WorkitemsListPostRequestDataItemAttributes(
            type=work_item.type,
            description=api_models.WorkitemsListPostRequestDataItemAttributesDescription(  # pylint: disable=line-too-long
                type=api_models.WorkitemsListPostRequestDataItemAttributesDescriptionType(  # pylint: disable=line-too-long
                    work_item.description.type
                ),
                value=work_item.description.value or oa_types.UNSET,
            ),
            status=work_item.status,
            title=work_item.title,
        )

        attrs.additional_properties.update(work_item.additional_attributes)

        return api_models.WorkitemsListPostRequestDataItem(
            type=api_models.WorkitemsListPostRequestDataItemType.WORKITEMS,
            attributes=attrs,
            # pylint: disable=line-too-long
            relationships=(
                api_models.WorkitemsListPostRequestDataItemRelationships(
                    module=api_models.WorkitemsListPostRequestDataItemRelationshipsModule(
                        data=api_models.WorkitemsListPostRequestDataItemRelationshipsModuleData(
                            id=f"{self._project_id}/{doc_ref.module_folder}/{doc_ref.module_name}",
                            type=api_models.WorkitemsListPostRequestDataItemRelationshipsModuleDataType.DOCUMENTS,
                        ),
                    )
                )
                if (doc_ref := work_item.home_document) is not None
                else oa_types.UNSET
            ),
            # pylint: enable=line-too-long
        )

    def _build_work_item_patch_request(
        self, work_item: dm.WorkItem
    ) -> api_models.WorkitemsSinglePatchRequest:
        attrs = api_models.WorkitemsSinglePatchRequestDataAttributes()
        if work_item.home_document:
            logger.warning(
                "Changing the work items home document is not supported."
            )

        if work_item.title is not None:
            attrs.title = work_item.title

        if work_item.description is not None:
            attrs.description = api_models.WorkitemsSinglePatchRequestDataAttributesDescription(  # pylint: disable=line-too-long
                type=api_models.WorkitemsSinglePatchRequestDataAttributesDescriptionType(  # pylint: disable=line-too-long
                    work_item.description.type
                ),
                value=work_item.description.value or oa_types.UNSET,
            )

        if work_item.status is not None:
            attrs.status = work_item.status

        attrs.additional_properties.update(work_item.additional_attributes)

        return api_models.WorkitemsSinglePatchRequest(
            data=api_models.WorkitemsSinglePatchRequestData(
                type=api_models.WorkitemsSinglePatchRequestDataType.WORKITEMS,
                id=f"{self._project_id}/{work_item.id}",
                attributes=attrs,
            )
        )

    def _post_work_item_batch(
        self,
        work_item_batch: api_models.WorkitemsListPostRequest,
        work_item_objs: list[dm.WorkItem],
    ):
        response = post_work_items.sync_detailed(
            self._project_id, client=self._client.client, body=work_item_batch
        )

        self._raise_on_error(response)

        assert (
            isinstance(response.parsed, api_models.WorkitemsListPostResponse)
            and response.parsed.data
        )
        for index, work_item_res in enumerate(response.parsed.data):
            assert work_item_res.id
            work_item_objs[index].id = work_item_res.id.split("/")[-1]

    def _calculate_post_work_item_request_sizes(
        self,
        work_item_data: api_models.WorkitemsListPostRequestDataItem,
        current_content_size: int = min_wi_request_size,
    ) -> t.Tuple[int, bool]:
        work_item_size = _get_json_content_size(work_item_data.to_dict())

        proj_content_size = current_content_size + work_item_size
        if current_content_size != min_wi_request_size:
            proj_content_size += len(b", ")

        return (
            proj_content_size,
            (work_item_size + min_wi_request_size)
            > self._client.max_content_size,
        )

    def _generate_work_item(
        self,
        work_item: (
            api_models.WorkitemsListGetResponseDataItem
            | api_models.WorkitemsSingleGetResponseData
        ),
        work_item_cls: type[WT],
    ) -> WT:
        assert work_item.attributes
        assert isinstance(work_item.id, str)
        work_item_id = work_item.id.split("/")[-1]
        links = []
        attachments = []
        home_document: dm.DocumentReference | None = None

        # We set both truncated flags to True and will only set them to False,
        # if the corresponding fields were requested and returned completely
        links_truncated = True
        attachments_truncated = True
        if work_item.relationships:
            if home_document_data := work_item.relationships.module:
                if home_document_data.data and home_document_data.data.id:
                    _, folder, name = home_document_data.data.id.split("/")
                    home_document = dm.DocumentReference(folder, name)

            if link_data := work_item.relationships.linked_work_items:
                if (
                    not link_data.meta
                    or link_data.meta.total_count is oa_types.UNSET
                ):
                    links_truncated = False

                links = [
                    self.links._parse_work_item_link(
                        link.id,
                        link.additional_properties.get("suspect", False),
                        work_item_id,
                    )
                    for link in link_data.data or []
                ]

            if attachment_data := work_item.relationships.attachments:
                if (
                    not attachment_data.meta
                    or attachment_data.meta.total_count is oa_types.UNSET
                ):
                    attachments_truncated = False

                attachments = [
                    dm.WorkItemAttachment(
                        work_item_id,
                        attachment.id.split("/")[-1],
                        None,  # title isn't provided
                    )
                    for attachment in attachment_data.data or []
                    if attachment.id
                ]

        description = None
        if work_item.attributes.description:
            description = dm.TextContent(
                self.unset_to_none(work_item.attributes.description.type),
                self.unset_to_none(work_item.attributes.description.value),
            )

        return work_item_cls(
            work_item_id,
            title=self.unset_to_none(work_item.attributes.title),
            description=description,
            type=self.unset_to_none(work_item.attributes.type),
            status=self.unset_to_none(work_item.attributes.status),
            additional_attributes=work_item.attributes.additional_properties,
            linked_work_items=links,
            attachments=attachments,
            linked_work_items_truncated=links_truncated,
            attachments_truncated=attachments_truncated,
            home_document=home_document,
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.linkedworkitems_list_post_request_data_item_relationships_work_item import (
        LinkedworkitemsListPostRequestDataItemRelationshipsWorkItem,
    )


T = TypeVar("T", bound="LinkedworkitemsListPostRequestDataItemRelationships")


@_attrs_define
class LinkedworkitemsListPostRequestDataItemRelationships:
    """
    Attributes:
        work_item (Union[Unset, LinkedworkitemsListPostRequestDataItemRelationshipsWorkItem]):
    """

    work_item: Union[
        Unset, "LinkedworkitemsListPostRequestDataItemRelationshipsWorkItem"
    ] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        work_item: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.work_item, Unset):
            work_item = self.work_item.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if work_item is not UNSET:
            field_dict["workItem"] = work_item

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.linkedworkitems_list_post_request_data_item_relationships_work_item import (
            LinkedworkitemsListPostRequestDataItemRelationshipsWorkItem,
        )

        d = src_dict.copy()
        _work_item = d.pop("workItem", UNSET)
        work_item: Union[
            Unset, LinkedworkitemsListPostRequestDataItemRelationshipsWorkItem
        ]
        if isinstance(_work_item, Unset):
            work_item = UNSET
        else:
            work_item = LinkedworkitemsListPostRequestDataItemRelationshipsWorkItem.from_dict(
                _work_item
            )

        linkedworkitems_list_post_request_data_item_relationships_obj = cls(
            work_item=work_item,
        )

        linkedworkitems_list_post_request_data_item_relationships_obj.additional_properties = (
            d
        )
        return linkedworkitems_list_post_request_data_item_relationships_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.workitems_list_get_response_data_item_relationships_approvals import (
        WorkitemsListGetResponseDataItemRelationshipsApprovals,
    )
    from ..models.workitems_list_get_response_data_item_relationships_assignee import (
        WorkitemsListGetResponseDataItemRelationshipsAssignee,
    )
    from ..models.workitems_list_get_response_data_item_relationships_attachments import (
        WorkitemsListGetResponseDataItemRelationshipsAttachments,
    )
    from ..models.workitems_list_get_response_data_item_relationships_author import (
        WorkitemsListGetResponseDataItemRelationshipsAuthor,
    )
    from ..models.workitems_list_get_response_data_item_relationships_backlinked_work_items import (
        WorkitemsListGetResponseDataItemRelationshipsBacklinkedWorkItems,
    )
    from ..models.workitems_list_get_response_data_item_relationships_categories import (
        WorkitemsListGetResponseDataItemRelationshipsCategories,
    )
    from ..models.workitems_list_get_response_data_item_relationships_comments import (
        WorkitemsListGetResponseDataItemRelationshipsComments,
    )
    from ..models.workitems_list_get_response_data_item_relationships_externally_linked_work_items import (
        WorkitemsListGetResponseDataItemRelationshipsExternallyLinkedWorkItems,
    )
    from ..models.workitems_list_get_response_data_item_relationships_linked_oslc_resources import (
        WorkitemsListGetResponseDataItemRelationshipsLinkedOslcResources,
    )
    from ..models.workitems_list_get_response_data_item_relationships_linked_revisions import (
        WorkitemsListGetResponseDataItemRelationshipsLinkedRevisions,
    )
    from ..models.workitems_list_get_response_data_item_relationships_linked_work_items import (
        WorkitemsListGetResponseDataItemRelationshipsLinkedWorkItems,
    )
    from ..models.workitems_list_get_response_data_item_relationships_module import (
        WorkitemsListGetResponseDataItemRelationshipsModule,
    )
    from ..models.workitems_list_get_response_data_item_relationships_planned_in import (
        WorkitemsListGetResponseDataItemRelationshipsPlannedIn,
    )
    from ..models.workitems_list_get_response_data_item_relationships_project import (
        WorkitemsListGetResponseDataItemRelationshipsProject,
    )
    from ..models.workitems_list_get_response_data_item_relationships_test_steps import (
        WorkitemsListGetResponseDataItemRelationshipsTestSteps,
    )
    from ..models.workitems_list_get_response_data_item_relationships_votes import (
        WorkitemsListGetResponseDataItemRelationshipsVotes,
    )
    from ..models.workitems_list_get_response_data_item_relationships_watches import (
        WorkitemsListGetResponseDataItemRelationshipsWatches,
    )
    from ..models.workitems_list_get_response_data_item_relationships_work_records import (
        WorkitemsListGetResponseDataItemRelationshipsWorkRecords,
    )


T = TypeVar("T", bound="WorkitemsListGetResponseDataItemRelationships")


@_attrs_define
class WorkitemsListGetResponseDataItemRelationships:
    """
    Attributes:
        approvals (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsApprovals]):
        assignee (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsAssignee]):
        attachments (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsAttachments]):
        author (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsAuthor]):
        backlinked_work_items (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsBacklinkedWorkItems]):
        categories (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsCategories]):
        comments (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsComments]):
        externally_linked_work_items (Union[Unset,
            WorkitemsListGetResponseDataItemRelationshipsExternallyLinkedWorkItems]):
        linked_oslc_resources (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsLinkedOslcResources]):
        linked_revisions (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsLinkedRevisions]):
        linked_work_items (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsLinkedWorkItems]):
        module (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsModule]):
        planned_in (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsPlannedIn]):
        project (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsProject]):
        test_steps (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsTestSteps]):
        votes (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsVotes]):
        watches (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsWatches]):
        work_records (Union[Unset, WorkitemsListGetResponseDataItemRelationshipsWorkRecords]):
    """

    approvals: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsApprovals"
    ] = UNSET
    assignee: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsAssignee"
    ] = UNSET
    attachments: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsAttachments"
    ] = UNSET
    author: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsAuthor"
    ] = UNSET
    backlinked_work_items: Union[
        Unset,
        "WorkitemsListGetResponseDataItemRelationshipsBacklinkedWorkItems",
    ] = UNSET
    categories: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsCategories"
    ] = UNSET
    comments: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsComments"
    ] = UNSET
    externally_linked_work_items: Union[
        Unset,
        "WorkitemsListGetResponseDataItemRelationshipsExternallyLinkedWorkItems",
    ] = UNSET
    linked_oslc_resources: Union[
        Unset,
        "WorkitemsListGetResponseDataItemRelationshipsLinkedOslcResources",
    ] = UNSET
    linked_revisions: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsLinkedRevisions"
    ] = UNSET
    linked_work_items: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsLinkedWorkItems"
    ] = UNSET
    module: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsModule"
    ] = UNSET
    planned_in: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsPlannedIn"
    ] = UNSET
    project: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsProject"
    ] = UNSET
    test_steps: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsTestSteps"
    ] = UNSET
    votes: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsVotes"
    ] = UNSET
    watches: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsWatches"
    ] = UNSET
    work_records: Union[
        Unset, "WorkitemsListGetResponseDataItemRelationshipsWorkRecords"
    ] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        approvals: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.approvals, Unset):
            approvals = self.approvals.to_dict()

        assignee: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.assignee, Unset):
            assignee = self.assignee.to_dict()

        attachments: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.attachments, Unset):
            attachments = self.attachments.to_dict()

        author: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.author, Unset):
            author = self.author.to_dict()

        backlinked_work_items: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.backlinked_work_items, Unset):
            backlinked_work_items = self.backlinked_work_items.to_dict()

        categories: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.categories, Unset):
            categories = self.categories.to_dict()

        comments: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.comments, Unset):
            comments = self.comments.to_dict()

        externally_linked_work_items: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.externally_linked_work_items, Unset):
            externally_linked_work_items = (
                self.externally_linked_work_items.to_dict()
            )

        linked_oslc_resources: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.linked_oslc_resources, Unset):
            linked_oslc_resources = self.linked_oslc_resources.to_dict()

        linked_revisions: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.linked_revisions, Unset):
            linked_revisions = self.linked_revisions.to_dict()

        linked_work_items: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.linked_work_items, Unset):
            linked_work_items = self.linked_work_items.to_dict()

        module: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.module, Unset):
            module = self.module.to_dict()

        planned_in: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.planned_in, Unset):
            planned_in = self.planned_in.to_dict()

        project: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.project, Unset):
            project = self.project.to_dict()

        test_steps: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.test_steps, Unset):
            test_steps = self.test_steps.to_dict()

        votes: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.votes, Unset):
            votes = self.votes.to_dict()

        watches: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.watches, Unset):
            watches = self.watches.to_dict()

        work_records: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.work_records, Unset):
            work_records = self.work_records.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if approvals is not UNSET:
            field_dict["approvals"] = approvals
        if assignee is not UNSET:
            field_dict["assignee"] = assignee
        if attachments is not UNSET:
            field_dict["attachments"] = attachments
        if author is not UNSET:
            field_dict["author"] = author
        if backlinked_work_items is not UNSET:
            field_dict["backlinkedWorkItems"] = backlinked_work_items
        if categories is not UNSET:
            field_dict["categories"] = categories
        if comments is not UNSET:
            field_dict["comments"] = comments
        if externally_linked_work_items is not UNSET:
            field_dict["externallyLinkedWorkItems"] = (
                externally_linked_work_items
            )
        if linked_oslc_resources is not UNSET:
            field_dict["linkedOslcResources"] = linked_oslc_resources
        if linked_revisions is not UNSET:
            field_dict["linkedRevisions"] = linked_revisions
        if linked_work_items is not UNSET:
            field_dict["linkedWorkItems"] = linked_work_items
        if module is not UNSET:
            field_dict["module"] = module
        if planned_in is not UNSET:
            field_dict["plannedIn"] = planned_in
        if project is not UNSET:
            field_dict["project"] = project
        if test_steps is not UNSET:
            field_dict["testSteps"] = test_steps
        if votes is not UNSET:
            field_dict["votes"] = votes
        if watches is not UNSET:
            field_dict["watches"] = watches
        if work_records is not UNSET:
            field_dict["workRecords"] = work_records

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.workitems_list_get_response_data_item_relationships_approvals import (
            WorkitemsListGetResponseDataItemRelationshipsApprovals,
        )
        from ..models.workitems_list_get_response_data_item_relationships_assignee import (
            WorkitemsListGetResponseDataItemRelationshipsAssignee,
        )
        from ..models.workitems_list_get_response_data_item_relationships_attachments import (
            WorkitemsListGetResponseDataItemRelationshipsAttachments,
        )
        from ..models.workitems_list_get_response_data_item_relationships_author import (
            WorkitemsListGetResponseDataItemRelationshipsAuthor,
        )
        from ..models.workitems_list_get_response_data_item_relationships_backlinked_work_items import (
            WorkitemsListGetResponseDataItemRelationshipsBacklinkedWorkItems,
        )
        from ..models.workitems_list_get_response_data_item_relationships_categories import (
            WorkitemsListGetResponseDataItemRelationshipsCategories,
        )
        from ..models.workitems_list_get_response_data_item_relationships_comments import (
            WorkitemsListGetResponseDataItemRelationshipsComments,
        )
        from ..models.workitems_list_get_response_data_item_relationships_externally_linked_work_items import (
            WorkitemsListGetResponseDataItemRelationshipsExternallyLinkedWorkItems,
        )
        from ..models.workitems_list_get_response_data_item_relationships_linked_oslc_resources import (
            WorkitemsListGetResponseDataItemRelationshipsLinkedOslcResources,
        )
        from ..models.workitems_list_get_response_data_item_relationships_linked_revisions import (
            WorkitemsListGetResponseDataItemRelationshipsLinkedRevisions,
        )
        from ..models.workitems_list_get_response_data_item_relationships_linked_work_items import (
            WorkitemsListGetResponseDataItemRelationshipsLinkedWorkItems,
        )
        from ..models.workitems_list_get_response_data_item_relationships_module import (
            WorkitemsListGetResponseDataItemRelationshipsModule,
        )
        from ..models.workitems_list_get_response_data_item_relationships_planned_in import (
            WorkitemsListGetResponseDataItemRelationshipsPlannedIn,
        )
        from ..models.workitems_list_get_response_data_item_relationships_project import (
            WorkitemsListGetResponseDataItemRelationshipsProject,
        )
        from ..models.workitems_list_get_response_data_item_relationships_test_steps import (
            WorkitemsListGetResponseDataItemRelationshipsTestSteps,
        )
        from ..models.workitems_list_get_response_data_item_relationships_votes import (
            WorkitemsListGetResponseDataItemRelationshipsVotes,
        )
        from ..models.workitems_list_get_response_data_item_relationships_watches import (
            WorkitemsListGetResponseDataItemRelationshipsWatches,
        )
        from ..models.workitems_list_get_response_data_item_relationships_work_records import (
            WorkitemsListGetResponseDataItemRelationshipsWorkRecords,
        )

        d = src_dict.copy()
        _approvals = d.pop("approvals", UNSET)
        approvals: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsApprovals
        ]
        if isinstance(_approvals, Unset):
            approvals = UNSET
        else:
            approvals = WorkitemsListGetResponseDataItemRelationshipsApprovals.from_dict(
                _approvals
            )

        _assignee = d.pop("assignee", UNSET)
        assignee: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsAssignee
        ]
        if isinstance(_assignee, Unset):
            assignee = UNSET
        else:
            assignee = WorkitemsListGetResponseDataItemRelationshipsAssignee.from_dict(
                _assignee
            )

        _attachments = d.pop("attachments", UNSET)
        attachments: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsAttachments
        ]
        if isinstance(_attachments, Unset):
            attachments = UNSET
        else:
            attachments = WorkitemsListGetResponseDataItemRelationshipsAttachments.from_dict(
                _attachments
            )

        _author = d.pop("author", UNSET)
        author: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsAuthor
        ]
        if isinstance(_author, Unset):
            author = UNSET
        else:
            author = (
                WorkitemsListGetResponseDataItemRelationshipsAuthor.from_dict(
                    _author
                )
            )

        _backlinked_work_items = d.pop("backlinkedWorkItems", UNSET)
        backlinked_work_items: Union[
            Unset,
            WorkitemsListGetResponseDataItemRelationshipsBacklinkedWorkItems,
        ]
        if isinstance(_backlinked_work_items, Unset):
            backlinked_work_items = UNSET
        else:
            backlinked_work_items = WorkitemsListGetResponseDataItemRelationshipsBacklinkedWorkItems.from_dict(
                _backlinked_work_items
            )

        _categories = d.pop("categories", UNSET)
        categories: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsCategories
        ]
        if isinstance(_categories, Unset):
            categories = UNSET
        else:
            categories = WorkitemsListGetResponseDataItemRelationshipsCategories.from_dict(
                _categories
            )

        _comments = d.pop("comments", UNSET)
        comments: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsComments
        ]
        if isinstance(_comments, Unset):
            comments = UNSET
        else:
            comments = WorkitemsListGetResponseDataItemRelationshipsComments.from_dict(
                _comments
            )

        _externally_linked_work_items = d.pop(
            "externallyLinkedWorkItems", UNSET
        )
        externally_linked_work_items: Union[
            Unset,
            WorkitemsListGetResponseDataItemRelationshipsExternallyLinkedWorkItems,
        ]
        if isinstance(_externally_linked_work_items, Unset):
            externally_linked_work_items = UNSET
        else:
            externally_linked_work_items = WorkitemsListGetResponseDataItemRelationshipsExternallyLinkedWorkItems.from_dict(
                _externally_linked_work_items
            )

        _linked_oslc_resources = d.pop("linkedOslcResources", UNSET)
        linked_oslc_resources: Union[
            Unset,
            WorkitemsListGetResponseDataItemRelationshipsLinkedOslcResources,
        ]
        if isinstance(_linked_oslc_resources, Unset):
            linked_oslc_resources = UNSET
        else:
            linked_oslc_resources = WorkitemsListGetResponseDataItemRelationshipsLinkedOslcResources.from_dict(
                _linked_oslc_resources
            )

        _linked_revisions = d.pop("linkedRevisions", UNSET)
        linked_revisions: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsLinkedRevisions
        ]
        if isinstance(_linked_revisions, Unset):
            linked_revisions = UNSET
        else:
            linked_revisions = WorkitemsListGetResponseDataItemRelationshipsLinkedRevisions.from_dict(
                _linked_revisions
            )

        _linked_work_items = d.pop("linkedWorkItems", UNSET)
        linked_work_items: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsLinkedWorkItems
        ]
        if isinstance(_linked_work_items, Unset):
            linked_work_items = UNSET
        else:
            linked_work_items = WorkitemsListGetResponseDataItemRelationshipsLinkedWorkItems.from_dict(
                _linked_work_items
            )

        _module = d.pop("module", UNSET)
        module: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsModule
        ]
        if isinstance(_module, Unset):
            module = UNSET
        else:
            module = (
                WorkitemsListGetResponseDataItemRelationshipsModule.from_dict(
                    _module
                )
            )

        _planned_in = d.pop("plannedIn", UNSET)
        planned_in: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsPlannedIn
        ]
        if isinstance(_planned_in, Unset):
            planned_in = UNSET
        else:
            planned_in = WorkitemsListGetResponseDataItemRelationshipsPlannedIn.from_dict(
                _planned_in
            )

        _project = d.pop("project", UNSET)
        project: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsProject
        ]
        if isinstance(_project, Unset):
            project = UNSET
        else:
            project = (
                WorkitemsListGetResponseDataItemRelationshipsProject.from_dict(
                    _project
                )
            )

        _test_steps = d.pop("testSteps", UNSET)
        test_steps: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsTestSteps
        ]
        if isinstance(_test_steps, Unset):
            test_steps = UNSET
        else:
            test_steps = WorkitemsListGetResponseDataItemRelationshipsTestSteps.from_dict(
                _test_steps
            )

        _votes = d.pop("votes", UNSET)
        votes: Union[Unset, WorkitemsListGetResponseDataItemRelationshipsVotes]
        if isinstance(_votes, Unset):
            votes = UNSET
        else:
            votes = (
                WorkitemsListGetResponseDataItemRelationshipsVotes.from_dict(
                    _votes
                )
            )

        _watches = d.pop("watches", UNSET)
        watches: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsWatches
        ]
        if isinstance(_watches, Unset):
            watches = UNSET
        else:
            watches = (
                WorkitemsListGetResponseDataItemRelationshipsWatches.from_dict(
                    _watches
                )
            )

        _work_records = d.pop("workRecords", UNSET)
        work_records: Union[
            Unset, WorkitemsListGetResponseDataItemRelationshipsWorkRecords
        ]
        if isinstance(_work_records, Unset):
            work_records = UNSET
        else:
            work_records = WorkitemsListGetResponseDataItemRelationshipsWorkRecords.from_dict(
                _work_records
            )

        workitems_list_get_response_data_item_relationships_obj = cls(
            approvals=approvals,
            assignee=assignee,
            attachments=attachments,
            author=author,
            backlinked_work_items=backlinked_work_items,
            categories=categories,
            comments=comments,
            externally_linked_work_items=externally_linked_work_items,
            linked_oslc_resources=linked_oslc_resources,
            linked_revisions=linked_revisions,
            linked_work_items=linked_work_items,
            module=module,
            planned_in=planned_in,
            project=project,
            test_steps=test_steps,
            votes=votes,
            watches=watches,
            work_records=work_records,
        )

        workitems_list_get_response_data_item_relationships_obj.additional_properties = (
            d
        )
        return workitems_list_get_response_data_item_relationships_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ProjecttemplatesListGetResponseLinks")


@_attrs_define
class ProjecttemplatesListGetResponseLinks:
    """
    Attributes:
        first (Union[Unset, str]):  Example: server-host-name/application-
            path/projecttemplates?page%5Bsize%5D=10&page%5Bnumber%5D=1.
        last (Union[Unset, str]):  Example: server-host-name/application-
            path/projecttemplates?page%5Bsize%5D=10&page%5Bnumber%5D=9.
        next_ (Union[Unset, str]):  Example: server-host-name/application-
            path/projecttemplates?page%5Bsize%5D=10&page%5Bnumber%5D=6.
        prev (Union[Unset, str]):  Example: server-host-name/application-
            path/projecttemplates?page%5Bsize%5D=10&page%5Bnumber%5D=4.
        self_ (Union[Unset, str]):  Example: server-host-name/application-
            path/projecttemplates?page%5Bsize%5D=10&page%5Bnumber%5D=5.
    """

    first: Union[Unset, str] = UNSET
    last: Union[Unset, str] = UNSET
    next_: Union[Unset, str] = UNSET
    prev: Union[Unset, str] = UNSET
    self_: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(
        init=False, factory=dict
    )

    def to_dict(self) -> Dict[str, Any]:
        first = self.first

        last = self.last

        next_ = self.next_

        prev = self.prev

        self_ = self.self_

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if first is not UNSET:
            field_dict["first"] = first
        if last is not UNSET:
            field_dict["last"] = last
        if next_ is not UNSET:
            field_dict["next"] = next_
        if prev is not UNSET:
            field_dict["prev"] = prev
        if self_ is not UNSET:
            field_dict["self"] = self_

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        first = d.pop("first", UNSET)

        last = d.pop("last", UNSET)

        next_ = d.pop("next", UNSET)

        prev = d.pop("prev", UNSET)

        self_ = d.pop("self", UNSET)

        projecttemplates_list_get_response_links_obj = cls(
            first=first,
            last=last,
            next_=next_,
            prev=prev,
            self_=self_,
        )

        projecttemplates_list_get_response_links_obj.additional_properties = d
        return projecttemplates_list_get_response_links_obj

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import copy
import email
import json
from email import message

import httpx
import pytest_httpx

import polarion_rest_api_client as polarion_api
from tests.conftest import (
    TEST_WIA_CREATED_RESPONSE,
    TEST_WIA_MULTI_CREATED_RESPONSE,
    TEST_WIA_NEXT_PAGE_RESPONSE,
    TEST_WIA_NO_NEXT_PAGE_RESPONSE,
)


def _extract_data_from_request(
    req: httpx.Request,
) -> dict[str, list[message.Message]]:
    headers = f"Content-Type: {req.headers['Content-Type']}\r\n\r\n"
    msg = email.message_from_bytes(
        headers.encode(req.headers.encoding) + req.content
    )
    fields: dict[str, list[message.Message]] = {}
    for part in msg.walk():
        field_name = part.get_param("name", header="Content-Disposition")
        if isinstance(field_name, str):
            fields.setdefault(field_name, []).append(part)
    return fields


def test_get_work_item_attachments_single_page(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(
        TEST_WIA_NO_NEXT_PAGE_RESPONSE,
        encoding="utf8",
    ) as f:
        httpx_mock.add_response(json=json.load(f))

    work_item_attachments = client.work_items.attachments.get_all(
        "MyWorkItemId",
        fields={"fields[workitem_attachments]": "id,title"},
    )

    query = {
        "fields[workitem_attachments]": "id,title",
        "page[size]": "100",
        "page[number]": "1",
    }
    reqs = httpx_mock.get_requests()
    assert reqs[0].method == "GET"
    assert dict(reqs[0].url.params) == query
    assert len(work_item_attachments) == 1
    assert len(reqs) == 1
    assert work_item_attachments[0] == polarion_api.WorkItemAttachment(
        "MyWorkItemId", "MyAttachmentId", "Title", file_name="File Name"
    )


def test_get_work_item_attachments_multi_page(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    with open(
        TEST_WIA_NEXT_PAGE_RESPONSE,
        encoding="utf8",
    ) as f:
        httpx_mock.add_response(json=json.load(f))
    with open(
        TEST_WIA_NO_NEXT_PAGE_RESPONSE,
        encoding="utf8",
    ) as f:
        httpx_mock.add_response(json=json.load(f))

    work_items_attachments = client.work_items.attachments.get_all(
        "MyWorkItemId"
    )

    query = {
        "fields[workitem_attachments]": "@basic",
        "page[size]": "100",
        "page[number]": "1",
    }
    reqs = httpx_mock.get_requests()
    assert len(reqs) == 2
    assert reqs[0].method == "GET"
    assert dict(reqs[0].url.params) == query
    assert reqs[1].method == "GET"
    query["page[number]"] = "2"
    assert dict(reqs[1].url.params) == query
    assert len(work_items_attachments) == 2


def test_delete_work_item_attachment(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
):
    httpx_mock.add_response(204)

    client.work_items.attachments.delete(
        polarion_api.WorkItemAttachment("MyWorkItemId", "Attachment", "Title")
    )

    req = httpx_mock.get_request()

    assert req is not None and req.method == "DELETE"
    assert req.url.path.endswith(
        "PROJ/workitems/MyWorkItemId/attachments/Attachment"
    )


def test_create_single_work_item_attachment(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
    work_item_attachment: polarion_api.WorkItemAttachment,
):
    with open(TEST_WIA_CREATED_RESPONSE, encoding="utf8") as f:
        httpx_mock.add_response(201, json=json.load(f))

    client.work_items.attachments.create(work_item_attachment)

    req = httpx_mock.get_request()
    fields = _extract_data_from_request(req)
    resource = fields["resource"][0]
    files = fields["files"]

    assert req is not None and req.method == "POST"
    assert req.url.path.endswith("PROJ/workitems/MyWorkItemId/attachments")
    assert work_item_attachment.id == "MyAttachmentId"
    assert work_item_attachment.content_bytes is not None

    assert resource is not None
    assert len(files) == 1

    assert resource.get_content_type() == "text/plain"
    assert resource.get_payload() == json.dumps(
        {
            "data": [
                {
                    "type": "workitem_attachments",
                    "attributes": {"fileName": "test.json", "title": "Title"},
                }
            ]
        }
    )

    assert files[0].get_content_type() == "text/plain"
    assert files[0].get_filename() == "test.json"
    assert files[0].get_payload() == work_item_attachment.content_bytes.decode(
        "utf-8"
    )


def test_create_multiple_work_item_attachments(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
    work_item_attachment: polarion_api.WorkItemAttachment,
):
    with open(TEST_WIA_MULTI_CREATED_RESPONSE, encoding="utf8") as f:
        httpx_mock.add_response(201, json=json.load(f))

    work_item_attachments = [
        work_item_attachment,
        copy.deepcopy(work_item_attachment),
        copy.deepcopy(work_item_attachment),
    ]

    client.work_items.attachments.create(work_item_attachments)

    req = httpx_mock.get_request()

    assert req is not None and req.method == "POST"
    assert req.url.path.endswith("PROJ/workitems/MyWorkItemId/attachments")

    assert work_item_attachments[0].id == "MyAttachmentId1"
    assert work_item_attachments[1].id == "MyAttachmentId2"
    assert work_item_attachments[2].id == "MyAttachmentId3"

    fields = _extract_data_from_request(req)
    resource = fields["resource"][0]
    files = fields["files"]

    assert len(files) == 3

    for index, file in enumerate(files):
        assert (
            content := work_item_attachments[index - 1].content_bytes
        ) is not None
        assert file.get_content_type() == "text/plain"
        assert file.get_filename() == "test.json"
        assert content.decode("utf-8")

    assert resource.get_content_type() == "text/plain"
    assert resource.get_payload() == json.dumps(
        {
            "data": 3
            * [
                {
                    "type": "workitem_attachments",
                    "attributes": {"fileName": "test.json", "title": "Title"},
                }
            ]
        }
    )


def test_update_work_item_attachment_title(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
    work_item_attachment: polarion_api.WorkItemAttachment,
):
    httpx_mock.add_response(204)

    work_item_attachment.id = "MyAttachmentId"
    work_item_attachment.title = "TestTitle"
    work_item_attachment.mime_type = None
    work_item_attachment.content_bytes = None
    work_item_attachment.file_name = None

    client.work_items.attachments.update(work_item_attachment)

    req = httpx_mock.get_request()

    fields = _extract_data_from_request(req)
    resource = fields["resource"][0]

    assert "files" not in fields
    assert resource.get_content_type() == "text/plain"
    assert resource.get_payload() == json.dumps(
        {
            "data": {
                "type": "workitem_attachments",
                "id": "PROJ/MyWorkItemId/MyAttachmentId",
                "attributes": {"title": "TestTitle"},
            }
        }
    )


def test_update_work_item_attachment_content(
    client: polarion_api.ProjectClient,
    httpx_mock: pytest_httpx.HTTPXMock,
    work_item_attachment: polarion_api.WorkItemAttachment,
):
    httpx_mock.add_response(204)

    work_item_attachment.id = "MyAttachmentId"
    work_item_attachment.title = None

    client.work_items.attachments.update(work_item_attachment)

    req = httpx_mock.get_request()

    fields = _extract_data_from_request(req)
    resource = fields["resource"][0]
    content = fields["content"][0]

    assert resource.get_content_type() == "text/plain"
    assert resource.get_payload() == json.dumps(
        {
            "data": {
                "type": "workitem_attachments",
                "id": "PROJ/MyWorkItemId/MyAttachmentId",
                "attributes": {},
            }
        }
    )

    assert content.get_content_type() == "text/plain"
    assert content.get_filename() == "test.json"
    assert work_item_attachment.content_bytes is not None
    assert content.get_payload() == work_item_attachment.content_bytes.decode(
        "utf-8"
    )
